"""
A shared caching layer for file-like objects.

pseudo_file objects can be used as drop-in replacements for actual file
handles to provide a transparent caching layer to avoid reading multiple
times from disk or network.

To create a pseudo_file instance, encapsulate a 'real' file handler
inside a lazy cache object:
  from dxtbx.filecache import lazy_file_cache
  cache = lazy_file_cache(open(filename, 'rb'))

Finally use a reference to the cache object to create one or many pseudo_file
instances:
  fh1 = cache.open()
  from dxtbx.filecache import pseudo_file
  fh2 = pseudo_file(cache) # equivalent
  fh3 = pseudo_file(cache)
  ...

Each pseudo_file instance can then be treated as a proper read-only
file handle, but will benefit from a shared cache:
  with cache.open() as fh:
    fh.read(100)
    fh.readline()
    fh.seek(500)
    fh.read()
    fh.readlines()
    fh.close()

To flush the cache and free the memory you can use
    cache.close()
This will drop the cache when all associated file handles are closed.
To instantly drop the cache you can use
    cache.force_close()
Any further access attempts will then result in an exception.
"""


from __future__ import annotations

import io
import os
from threading import Lock


class lazy_file_cache:
    """An object providing shared cached access to files"""

    def __init__(self, file_object):
        """Create a shared cache based on a single file handle."""
        # Uncomment this line to print debug information
        # self._debug = self._debug_enable

        # Size of a block to read. This should not be smaller than 4k, which is the
        # default block size on many systems.
        self._page_size = max(4096, io.DEFAULT_BUFFER_SIZE)

        # Reference to the underlying file object. When no further information can
        # be gained from the file (ie. it has been read once completely), it may
        # be closed.
        self._all_cached = False
        self._file = file_object
        self._file_lock = Lock()

        # BytesIO object containing cached information
        self._cache_object = io.BytesIO()
        self._cache_size = 0
        self._cache_limit = 4 * 1024 * 1024
        self._cache_limit_reached = False

        # Current status of lazy cache towards client objects.
        # Opening new file handles is disallowed when the object is closing.
        self._closing = False
        # When the lazy cache object is closed no further access is allowed,
        # and cached information is dropped.
        self._closed = False

        # Number of currently registered client objects
        self._reference_counter = 0
        self._reference_counter_lock = Lock()

        self._debug("Created cache object for %s: %s" % (str(file_object), str(self)))

    def _debug(self, string):
        pass

    def _debug_enable(self, string):
        print("%s: %s" % (format(id(self), "#x"), string))

    def __del__(self):
        """Close file handles and drop cache on garbage collection."""
        try:
            self.force_close()
        except TypeError:
            # Ignore errors that might happen during interpreter shutdown
            pass

    def _cache_up_to(self, position):
        '''Ensure that the file has been read up to "position"'''

        # Limit reads to upper cache size limit
        if position > self._cache_limit:
            position = self._cache_limit

        # Is read actually necessary?
        if self._all_cached or (position < self._cache_size):
            return

        with self._file_lock:
            read_bytes = position - self._cache_size
            # This looks like an unnecessary check, but required for concurrency
            if read_bytes <= 0:
                return

            # Do not read less than a memory page, round up read size to a
            # multiple of page sizes if necessary.
            read_bytes = self._page_size * (
                (read_bytes + self._page_size - 1) // self._page_size
            )
            self._debug("Reading %d bytes from file" % read_bytes)

            expected_cache_size = self._cache_size + read_bytes

            data = self._file.read(read_bytes)
            self._cache_object.seek(self._cache_size)
            self._cache_object.write(data)
            self._cache_size = self._cache_object.tell()

            self._debug(
                "Read %d bytes from file, cache size %d" % (len(data), self._cache_size)
            )

            if expected_cache_size != self._cache_size:
                # must have reached end of file
                self._debug(
                    "Lazy cache reached EOF (%d != %d)"
                    % (expected_cache_size, self._cache_size)
                )
                self._all_cached = True
                self._close_file()

            if self._cache_size >= self._cache_limit:
                self._cache_limit_reached = True
                self._debug("Cache limit reached with %d bytes" % self._cache_size)

    def _cache_all(self):
        """Read entire remaining file into cache."""

        # Is read actually necessary?
        if self._all_cached:
            return

        # Is read actually possible?
        if self._cache_limit_reached:
            return

        with self._file_lock:
            # Check again with lock held, required for concurrency
            if self._all_cached or self._cache_limit_reached:
                return

            self._debug("Reading remaining file into cache")

            data = self._file.read(self._cache_limit - self._cache_size)
            self._cache_object.seek(self._cache_size)
            self._cache_object.write(data)
            self._debug("Read %d bytes" % len(data))
            self._cache_size += len(data)

            if self._cache_size >= self._cache_limit:
                # Don't cache more than the set limit. In this case keep file handler
                # open and pass read requests through.
                self._cache_limit_reached = True
                self._debug("Cache size limit reached")
                return

            self._all_cached = True
            self._close_file()

    def _check_not_closed(self):
        if self._closed:
            self._debug("Instance tried to access closed cache")
            raise OSError(
                "Accessing lazy file cache %s after closing is not allowed" % str(self)
            )

    def _close_file(self):
        if self._file is not None:
            self._debug(
                "Closing lazy cache internal file handle (%d bytes read)"
                % self._cache_size
            )
            self._file.close()

    def open(self):
        """Create and return a new pseudo_file object for this cache."""
        return pseudo_file(self)

    def close(self):
        if not self._closing:
            self._closing = True
            self._debug("Closing lazy cache %s" % str(self))
            with self._reference_counter_lock:
                if self._reference_counter == 0:
                    self.force_close()

    def force_close(self):
        """Close encapsulated file handle, drop cache and prevent further reads."""
        if not self._closed:
            self._closing = True
            self._closed = True
            self._close_file()
            if self._reference_counter > 0:
                self._debug(
                    "Warning: %d connected instances remain" % self._reference_counter
                )
            if self._cache_object is not None:
                self._cache_object.close()
                self._cache_object = None

    def register(self):
        """Register a client object. Reference counting for debug purposes."""
        with self._reference_counter_lock:
            self._check_not_closed()
            if self._closing:
                self._debug("Instance tried to connect to closing cache")
                raise OSError("Cannot open new file handle: lazy file cache is closing")
            self._reference_counter += 1
        self._debug("Instance connected to lazy cache")

    def unregister(self):
        """Unregister a client object. Reference counting for debug purposes."""
        self._debug("Instance disconnected from lazy cache")
        with self._reference_counter_lock:
            self._reference_counter -= 1
            if self._closing and (self._reference_counter == 0):
                self.force_close()

    def pass_read(self, start=0, maxbytes=None):
        """Read from position start up to maxbytes bytes from file.
        If maxbytes is not set, read the entire file."""
        self._check_not_closed()

        # Do we need to pass the read request to the underlying file object?
        passthrough = False

        if not self._all_cached:
            # Ensure that relevant data is in cache
            if maxbytes is None:
                self._cache_all()
                if not self._all_cached:
                    passthrough = True
            else:
                self._cache_up_to(start + maxbytes)
                if self._cache_limit_reached:
                    passthrough = start + maxbytes > self._cache_size

        if passthrough:
            with self._file_lock:
                self._file.seek(start)
                if maxbytes is None:
                    self._debug("Passing through read from %d" % start)
                    return self._file.read(), self._file.tell()
                else:
                    self._debug(
                        "Passing through read from %d to %d"
                        % (start, start + maxbytes - 1)
                    )
                    return self._file.read(maxbytes), self._file.tell()
        else:
            self._cache_object.seek(start)
            if maxbytes is None:
                return self._cache_object.read(), self._cache_object.tell()
            else:
                return self._cache_object.read(maxbytes), self._cache_object.tell()

    def pass_readline(self, start=0, maxbytes=None):
        """Read a line from file, but no more than maxbytes bytes."""
        self._check_not_closed()

        if self._all_cached:
            self._cache_object.seek(start)
            if maxbytes is None:
                return self._cache_object.readline(), self._cache_object.tell()
            else:
                return self._cache_object.readline(maxbytes), self._cache_object.tell()

        if self._cache_size <= start:
            self._cache_up_to(start + self._page_size)

        self._cache_object.seek(start)
        if maxbytes is None:
            line_candidate = self._cache_object.readline()
        else:
            line_candidate = self._cache_object.readline(maxbytes)

        end_position = self._cache_object.tell()

        if end_position < self._cache_size or line_candidate.endswith(b"\n"):
            # Found a complete line within the cache
            return line_candidate, end_position

        if (maxbytes is not None) and (end_position == start + maxbytes):
            # Fulfilled maxbytes condition within the cache
            return line_candidate, end_position

        # Need more data
        while end_position == self._cache_size and not self._cache_limit_reached:
            # Do we have a complete line?
            if line_candidate.endswith(b"\n"):
                return line_candidate, end_position

            # Ran against cache limit. Extend cache
            self._cache_up_to(self._cache_size + self._page_size)
            self._cache_object.seek(end_position)

            # Continue reading
            if maxbytes is None:
                line_candidate += self._cache_object.readline()
            else:
                foundbytes = end_position - start
                line_candidate += self._cache_object.readline(maxbytes - foundbytes)
            end_position = self._cache_object.tell()

        # Do we have a complete line?
        if line_candidate.endswith(b"\n") or self._all_cached:
            return line_candidate, end_position

        assert self._cache_limit_reached  # Only legitimate way of reaching here

        with self._file_lock:
            self._file.seek(end_position)
            if maxbytes is None:
                return line_candidate + self._file.readline(), self._file.tell()
            else:
                foundbytes = end_position - start
                return (
                    line_candidate + self._file.readline(maxbytes - foundbytes),
                    self._file.tell(),
                )


class pseudo_file:
    """A file-like object that serves as frontend to a dxtbx lazy file cache."""

    def __init__(self, lazy_cache_object):
        self._cache_object = lazy_cache_object
        self._cache_object.register()
        self._closed = False
        self._seek = 0

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __iter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def _check_not_closed(self):
        if self._closed:
            raise OSError("Accessing lazy file cache after closing is not allowed")

    def close(self):
        self._closed = True
        if self._cache_object is not None:
            self._cache_object.unregister()
            self._cache_object = None

    def flush(self):
        self._check_not_closed()

    def __next__(self):
        data = self.readline()
        if data == b"":
            raise StopIteration()
        return data

    def readable(self):
        return not self._closed

    def writable(self):
        return False

    def seekable(self):
        return self.readable()

    @property
    def closed(self):
        return self._closed

    def read(self, size=-1):
        self._check_not_closed()
        if size > 0:
            data, self._seek = self._cache_object.pass_read(
                start=self._seek, maxbytes=size
            )
        elif size == 0:
            data = b""
        else:
            data, self._seek = self._cache_object.pass_read(start=self._seek)
        return data

    def readline(self, size=-1):
        self._check_not_closed()
        if size > 0:
            data, self._seek = self._cache_object.pass_readline(
                start=self._seek, maxbytes=size
            )
        else:
            data, self._seek = self._cache_object.pass_readline(start=self._seek)
        return data

    def readlines(self, sizehint=-1):
        self._check_not_closed()
        if sizehint > 0:
            data, self._seek = self._cache_object.pass_read(
                start=self._seek, maxbytes=sizehint
            )
        else:
            data, self._seek = self._cache_object.pass_read(start=self._seek)
        return data.splitlines(True)

    def seek(self, offset, whence=os.SEEK_SET):
        self._check_not_closed()
        if whence == os.SEEK_SET:
            self._seek = offset
        elif whence == os.SEEK_CUR:
            self._seek += offset
        else:
            raise NotImplementedError(
                "Seeking relative to file length is not supported"
            )

    def tell(self):
        self._check_not_closed()
        return self._seek

    def truncate(self, size=0):
        raise NotImplementedError("Truncating lazy file caches is not allowed")

    def write(self, string):
        raise NotImplementedError("Writing to lazy file caches is not allowed")

    def writelines(self, sequence):
        raise NotImplementedError("Writing to lazy file caches is not allowed")



from __future__ import annotations

import dxtbx.data.beamline_defs as dxbd


def test_lookup_unknown_detector():
    n = dxbd.get_beamline_definition(
        "This is a detector serial number that does not exist"
    )
    assert "Dummy CIF" in str(n), n
    assert str(n.CIF_block()) == ""
    assert n.CIF_block().__module__ == "iotbx.cif.model"
    assert str(n.mmCIF_block()) == ""
    assert n.mmCIF_block().__module__ == "iotbx.cif.model"


def test_lookup_known_detector():
    n = dxbd.get_beamline_definition("PILATUS 2M, S/N 24-0107 Diamond")
    assert "Dummy" not in str(n), n
    assert str(n) != ""
    cif = n.CIF_block()
    assert cif.__module__ == "iotbx.cif.model"
    assert "_diffrn_source" in str(cif)
    mmcif = n.mmCIF_block()
    assert mmcif.__module__ == "iotbx.cif.model"
    assert "_diffrn_source.source" in str(mmcif)



from __future__ import annotations

import binascii
import math
from pathlib import Path

import h5py
import hdf5plugin  # noqa; F401
import numpy as np
import nxmx
import pint
from tqdm import tqdm

import dxtbx.model
import dxtbx.nexus
from dxtbx.ext import compress


def compute_cbf_header(nxmx_obj: nxmx.NXmx, nn: int):
    nxentry = nxmx_obj.entries[0]
    nxsample = nxentry.samples[0]
    nxinstrument = nxentry.instruments[0]
    nxdetector = nxinstrument.detectors[0]
    nxbeam = nxinstrument.beams[0]

    beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
    wavelength = beam_factory.make_beam(index=0).get_wavelength()
    distance = nxdetector.distance
    if distance is None:
        distance = pint.Quantity(
            dxtbx.nexus.get_dxtbx_detector(nxdetector, wavelength)[0].get_distance(),
            "mm",
        )

    result = []

    A = nxinstrument.attenuators[0]["attenuator_transmission"][()]

    # this timestamp _should_ be in UTC - so can ignore timezone info - at
    # least for purposes here (causes errors for old versions of dxtbx reading
    # the data) - 19 chars needed
    timestamp = nxentry.start_time

    dependency_chain = nxmx.get_dependency_chain(nxsample.depends_on)

    result.append("###CBF: VERSION 1.5, CBFlib v0.7.8 - Eiger detectors")
    result.append("")
    result.append(f"data_{nn + 1:06d}")
    result.append("")
    result.append(
        """_array_data.header_convention "PILATUS_1.2"
_array_data.header_contents
;"""
    )
    result.append(
        f"# Detector: {nxdetector.description.upper()} S/N {nxdetector.serial_number} "
        f"{nxinstrument.short_name}"
    )
    result.append(f"# {timestamp}")
    px_size_fast = (
        nxdetector.modules[0].fast_pixel_direction[()].to("m").magnitude.item()
    )
    px_size_slow = (
        nxdetector.modules[0].slow_pixel_direction[()].to("m").magnitude.item()
    )
    result.append(f"# Pixel_size {px_size_fast} m x {px_size_slow} m")
    result.append(
        f"# {nxdetector.sensor_material} sensor, "
        f"thickness {nxdetector.sensor_thickness:~}"
    )
    result.append(f"# Exposure_time {nxdetector.count_time.to('s'):~.5f}")
    result.append(f"# Exposure_period {nxdetector.count_time.to('s'):~.5f}")
    result.append("# Tau = 1e-9 s")
    result.append(f"# Count_cutoff {nxdetector.saturation_value} counts")
    result.append("# Threshold_setting: 0 eV")
    result.append("# Gain_setting: mid gain (vrf = -0.200)")
    result.append("# N_excluded_pixels = 0")
    result.append(
        f"# Wavelength {nxbeam.incident_wavelength.to('angstrom').magnitude:.5f} A"
    )
    result.append(f"# Detector_distance {distance.to('m').magnitude:.5f} m")
    result.append(
        f"# Beam_xy ({nxdetector.beam_center_x.magnitude:.2f}, {nxdetector.beam_center_y.magnitude:.2f}) "
        "pixels"
    )
    result.append("# Flux 0.000000")
    result.append(f"# Filter_transmission {A:.3f}")

    transformations = {
        t.path.split("/")[-1].capitalize(): t for t in nxinstrument.transformations
    }
    two_theta = transformations.get("TWO_THETA", np.array(0))[()]
    result.append(f"# Detector_2theta {two_theta:.3f} deg.")

    result.append("# Polarization 0.990")

    rotations = {
        t.path.split("/")[-1].capitalize(): t
        for t in dependency_chain
        if t.transformation_type == "rotation"
    }

    if "PHI" in rotations and "KAPPA" in rotations:
        kappa_axis = rotations["KAPPA"].attrs["vector"]
        phi_axis = rotations["PHI"].attrs["vector"]
        alpha = np.degrees(np.arccos(np.dot(phi_axis, kappa_axis)))
    else:
        alpha = 0
    result.append(f"# Alpha {alpha:.4f} deg.")

    for name, transformation in rotations.items():
        if len(transformation) > 1 and transformation.transformation_type == "rotation":
            image_start_angle = float(transformation[nn].to("degree").magnitude)
            if transformation.end:
                angle_increment = (
                    float(transformation.end[nn].to("degree").magnitude)
                    - image_start_angle
                )
            else:
                # Fallback to inferring the per-image oscillation from the image start angles
                angle_increment = float(
                    (transformation[1] - transformation[0]).to("degree").magnitude
                )
            if angle_increment:
                result.append(f"# Start_angle {image_start_angle:.4f} deg.")
                result.append(f"# Angle_increment {angle_increment:.4f} deg.")
        else:
            image_start_angle = float(transformation[0].to("degree").magnitude)
            angle_increment = 0
        result.append(f"# {name} {image_start_angle:.4f} deg.")
        result.append(f"# {name}_increment {angle_increment:.4f} deg.")

    result.append("# Oscillation_axis X.CW")
    result.append("# N_oscillations 1")
    result.append(";")

    return "\n".join(result)


def make_cbf(
    in_name: Path | str,
    output_directory: Path,
    template: str | None = None,
    num_digits: int = 0,
):
    with h5py.File(in_name) as f:
        start_tag = binascii.unhexlify("0c1a04d5")

        nxmx_obj = nxmx.NXmx(f)
        nxsample = nxmx_obj.entries[0].samples[0]
        nxinstrument = nxmx_obj.entries[0].instruments[0]
        nxdetector = nxinstrument.detectors[0]
        nxdata = nxmx_obj.entries[0].data[0]

        dependency_chain = nxmx.get_dependency_chain(nxsample.depends_on)
        scan_axis = None
        for t in dependency_chain:
            # Find the first varying rotation axis
            if (
                t.transformation_type == "rotation"
                and len(t) > 1
                and not np.all(t[()] == t[0])
            ):
                scan_axis = t
                break

        if scan_axis is None:
            # Fall back on the first varying axis of any type
            for t in dependency_chain:
                if len(t) > 1 and not np.all(t[()] == t[0]):
                    scan_axis = t
                    break

        if scan_axis is None:
            scan_axis = nxsample.depends_on

        static_mask = dxtbx.nexus.get_static_mask(nxdetector)
        bit_depth_readout = nxdetector.bit_depth_readout

        if template is None:
            template = Path(in_name).stem + "_"
        num_images = len(scan_axis)
        num_digits = max(num_digits, int(math.log10(num_images)) + 1)

        print(f"Writing images to {template}{'#' * num_digits}.cbf:")
        for j in tqdm(range(num_images), unit=" images"):
            header = compute_cbf_header(nxmx_obj, j)
            (data,) = dxtbx.nexus.get_raw_data(nxdata, nxdetector, j)
            if bit_depth_readout:
                # if 32 bit then it is a signed int, I think if 8, 16 then it is
                # unsigned with the highest two values assigned as masking values
                if bit_depth_readout == 32:
                    top = 2**31
                else:
                    top = 2**bit_depth_readout
                d1d = data.as_1d()
                d1d.set_selected(d1d == top - 1, -1)
                d1d.set_selected(d1d == top - 2, -2)

            # set the tile join regions to -1 - MOSFLM cares about this apparently
            if static_mask:
                data.as_1d().set_selected(~static_mask[0].as_1d(), -1)
            compressed = compress(data)

            mime = f"""

_array_data.data
;
--CIF-BINARY-FORMAT-SECTION--
Content-Type: application/octet-stream;
     conversions="x-CBF_BYTE_OFFSET"
Content-Transfer-Encoding: BINARY
X-Binary-Size: {len(compressed)}
X-Binary-ID: 1
X-Binary-Element-Type: "signed 32-bit integer"
X-Binary-Element-Byte-Order: LITTLE_ENDIAN
X-Binary-Number-of-Elements: {data.size()}
X-Binary-Size-Fastest-Dimension: {data.focus()[1]}
X-Binary-Size-Second-Dimension: {data.focus()[0]}
X-Binary-Size-Padding: 4095

"""

            padding = (
                bytearray(4095)
                + b"""--CIF-BINARY-FORMAT-SECTION----
;"""
            )

            filename = f"{template}{j + 1:0{num_digits}d}.cbf"
            with open(output_directory / filename, "wb") as fout:
                fout.write(
                    ("".join(header) + mime).replace("\n", "\r\n").encode("latin-1")
                )
                fout.write(start_tag)
                fout.write(compressed)
                fout.write(padding)



"""An implementation of the CBF image reader for Pilatus images, for the P12M-DLS"""


from __future__ import annotations

import binascii
import calendar
import math
import sys

import libtbx
from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus
from dxtbx.format.FormatCBFMiniPilatusHelpers import get_pilatus_timestamp
from dxtbx.masking import GoniometerMaskerFactory
from dxtbx.model import Detector, ParallaxCorrectedPxMmStrategy


class FormatCBFMiniPilatusDLS12M(FormatCBFMiniPilatus):
    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 120-0100" in header
            ):
                return True

        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return True
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        # if multi_panel == False, then interpret data as 24 panels, where each
        # row of 5 panels is grouped as one "panel"
        # elif multi_panel == True, then interpret data as 120 panels,
        # 24 rows * 5 columns
        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        self._multi_panel = kwargs.get("multi_panel", False)

        super().__init__(image_file, **kwargs)

    def _detector(self):

        # module positions from detector blueprints - modelling at the moment as
        # 24 modules, each consisting of 5 sensors (the latter is ignored)
        x = matrix.col((-1, 0, 0))
        y = matrix.col((0, 1, 0))
        z = matrix.col((0, 0, 1))

        beam_xy = self._cif_header_dictionary["Beam_xy"]
        beam_xy = beam_xy.replace("(", "").replace(")", "").replace(",", "").split()[:2]
        obs_beam_x, obs_beam_y = [float(f) for f in beam_xy]

        ideal_beam_x = 1075
        ideal_beam_y = 2594

        beam_shift_x = 0.172 * (ideal_beam_x - obs_beam_x)
        beam_shift_y = 0.172 * (ideal_beam_y - obs_beam_y)

        distance = (
            float(self._cif_header_dictionary["Detector_distance"].split()[0]) * 1000.0
        )

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        thickness = float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0

        off_x = 184.9

        detector = Detector()
        root = detector.hierarchy()
        root.set_frame(
            x.elems,
            y.elems,
            (-distance * z + (beam_shift_x * x) + (beam_shift_y * y)).elems,
        )

        table = attenuation_coefficient.get_table("Si")
        mu = table.mu_at_angstrom(wavelength) / 10.0
        t0 = thickness
        px_mm = ParallaxCorrectedPxMmStrategy(mu, t0)

        self.coords = {}

        for j in range(24):
            shift_y = 195 + 17
            ymin, ymax = j * shift_y, j * shift_y + 195

            angle = math.pi * (-12.2 + 0.5 * 7.903 + j * (7.903 + 0.441)) / 180.0
            fast = matrix.col((1, 0, 0))
            slow = matrix.col((0, math.sin(angle), math.cos(angle)))
            normal = fast.cross(slow)

            row_origin = 250.0 * normal - off_x * fast - 16.8 * slow

            if not self._multi_panel:
                xmin, xmax = 0, 2463

                # OK two calls to add_panel here for detector like things => two
                # copies of the panel then? https://github.com/dials/dials/issues/189
                # ... this is also not the source of the leak

                # OBS! you need to set the panel to a root before set local frame...
                p = root.add_panel()
                p.set_type("SENSOR_PAD")
                p.set_name("row-%02d" % j)
                p.set_raw_image_offset((xmin, ymin))
                p.set_image_size((2463, 195))
                p.set_trusted_range((0, 1000000))
                p.set_pixel_size((0.172, 0.172))
                p.set_local_frame(fast.elems, slow.elems, row_origin.elems)
                p.set_thickness(thickness)
                p.set_material("Si")
                p.set_mu(mu)
                p.set_px_mm_strategy(px_mm)
                p.set_raw_image_offset((xmin, ymin))
                p.set_projection_2d((1, 0, 0, 1), (-j * (195 + 17), 0))
                self.coords[p.get_name()] = (xmin, ymin, xmax, ymax)

            else:
                shift_x = 487 + 7

                for i in range(5):
                    xmin, xmax = i * shift_x, i * shift_x + 487
                    origin = row_origin + i * (487 + 7) * 0.172 * fast

                    # OBS! you need to set the panel to a root before set local frame...
                    p = root.add_panel()
                    p.set_type("SENSOR_PAD")
                    p.set_name("row-%02d-col-%02d" % (j, i))
                    p.set_raw_image_offset((xmin, ymin))
                    p.set_image_size((487, 195))
                    p.set_trusted_range((0, 1000000))
                    p.set_pixel_size((0.172, 0.172))
                    p.set_local_frame(fast.elems, slow.elems, origin.elems)
                    p.set_thickness(thickness)
                    p.set_material("Si")
                    p.set_mu(mu)
                    p.set_px_mm_strategy(px_mm)
                    p.set_raw_image_offset((xmin, ymin))
                    p.set_projection_2d((1, 0, 0, 1), (-j * (195 + 17), -i * (487 + 7)))
                    self.coords[p.get_name()] = (xmin, ymin, xmax, ymax)

        detector = self._mask_bad_modules(detector)
        return detector

    def _mask_bad_modules(self, detector):
        # Mask out known bad modules
        timestamp = get_pilatus_timestamp(self._cif_header_dictionary["timestamp"])
        # https://photon-science.desy.de/research/technical_groups/detectors/e190302/e199162/CharacterizationandCalibrationofPilatus.pdf
        nx = 487  # module pixels x
        ny = 195  # module pixels y
        cx = 60  # chip pixels x
        cy = 97  # chip pixels y
        dx = 7  # module gap size

        if timestamp > calendar.timegm((2022, 7, 1, 0, 0, 0)):
            # 2022 run 3
            # Detector serviced by Dectris, no bad modules
            pass
        elif timestamp > calendar.timegm((2022, 1, 24, 0, 0, 0)):
            # 2022 run 1
            # module @ row 10 column 3
            if self._multi_panel:
                detector[5 * 10 + 3].add_mask(0, 0, nx, ny)
            else:
                detector[10].add_mask((nx + dx) * 3, 0, (nx + dx) * 3 + nx, ny)
        elif timestamp > calendar.timegm((2020, 9, 8, 0, 0, 0)):
            # 2020 run 4
            # Detector serviced by Dectris, no bad modules
            pass
        elif timestamp > calendar.timegm((2020, 2, 21, 0, 0, 0)):
            # 2020 run 1
            # module @ row 1 column 4
            # blank chip @ row 15 column 4 (chip row 0, column 1)
            # module @ row 17 column 0
            # module @ row 17 column 4
            if self._multi_panel:
                detector[5 * 1 + 4].add_mask(0, 0, nx, ny)
                detector[5 * 17].add_mask(0, 0, nx, ny)
                detector[5 * 17 + 4].add_mask(0, 0, nx, ny)
                detector[5 * 15 + 4].add_mask(cx, 0, cx * 2, cy)
            else:
                detector[1].add_mask((nx + dx) * 4, 0, (nx + dx) * 4 + nx, ny)
                detector[17].add_mask(0, 0, nx, ny)
                detector[17].add_mask((nx + dx) * 4, 0, (nx + dx) * 4 + nx, ny)
                detector[15].add_mask((nx + dx) * 4 + cx, 0, (nx + dx) * 4 + cx * 2, cy)
        elif timestamp > calendar.timegm((2019, 11, 26, 0, 0, 0)):
            # 2019 run 5
            # module @ row 17 column 0
            # module @ row 17 column 4
            if self._multi_panel:
                detector[5 * 17].add_mask(0, 0, nx, ny)
                detector[5 * 17 + 4].add_mask(0, 0, nx, ny)
            else:
                detector[17].add_mask(0, 0, nx, ny)
                detector[17].add_mask((nx + dx) * 4, 0, (nx + dx) * 4 + nx, ny)
        elif timestamp > calendar.timegm((2019, 9, 3, 0, 0, 0)):
            # 2019 run 4
            # module @ row 15 column 2
            # module @ row 17 column 0
            if self._multi_panel:
                detector[5 * 15 + 2].add_mask(0, 0, nx, ny)
                detector[5 * 17].add_mask(0, 0, nx, ny)
            else:
                detector[15].add_mask((nx + dx) * 2, 0, (nx + dx) * 2 + nx, ny)
                detector[17].add_mask(0, 0, nx, ny)

        return detector

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values

    def get_raw_data(self):
        if self._raw_data is None:
            raw_data = self._read_cbf_image()
            self._raw_data = []

            for panel in self.get_detector():
                xmin, ymin = panel.get_raw_image_offset()
                xmax = xmin + panel.get_image_size()[0]
                ymax = ymin + panel.get_image_size()[1]
                self._raw_data.append(raw_data[ymin:ymax, xmin:xmax])

        return tuple(self._raw_data)

    def get_goniometer_shadow_masker(self, goniometer=None):
        if goniometer is None:
            goniometer = self.get_goniometer()
        return GoniometerMaskerFactory.dls_i23_kappa(goniometer)

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header."""

        timestamp = get_pilatus_timestamp(self._cif_header_dictionary["timestamp"])
        # Goniometer changed from reverse phi to conventional rotation direction
        # on this date:
        # calendar.timegm(time.strptime('2016-04-01T00:00:00', '%Y-%m-%dT%H:%M:%S'))
        if timestamp < 1459468800:
            return self._goniometer_factory.single_axis_reverse()

        alpha = 50.0
        if "Phi" in self._cif_header_dictionary:
            phi_value = float(self._cif_header_dictionary["Phi"].split()[0])
        else:
            phi_value = 0.0

        if "Kappa" in self._cif_header_dictionary:
            kappa_value = float(self._cif_header_dictionary["Kappa"].split()[0])
        else:
            kappa_value = 0.0

        if "Omega" in self._cif_header_dictionary:
            omega_value = float(self._cif_header_dictionary["Omega"].split()[0])
        else:
            omega_value = 0.0

        return self._goniometer_factory.make_kappa_goniometer(
            alpha, omega_value, kappa_value, phi_value, "-y", "omega"
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDLS12M.understand(arg))



# LIBTBX_SET_DISPATCHER_NAME dxtbx.show_mask_info

from __future__ import annotations

import argparse
import sys

import dxtbx.util
from dxtbx.model.experiment_list import ExperimentListFactory


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser()
    parser.add_argument("filenames", metavar="IMAGE", nargs="+")
    options = parser.parse_args(args)
    try:
        el = ExperimentListFactory.from_filenames(options.filenames)
    except FileNotFoundError as e:
        sys.exit(str(e))

    dxtbx.util.show_mask_info(el)


if __name__ == "__main__":
    run()



"""Classes for Formats that recognise data stored in the MRC format, an open
standard used in electron microscopy
(http://www.ccpem.ac.uk/mrc_format/mrc2014.php)"""


from __future__ import annotations

import logging
import os
import re
from math import sqrt

import mrcfile
from mrcfile.constants import MAP_ID, MAP_ID_OFFSET_BYTES

from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.Format import Format
from dxtbx.format.FormatMultiImage import FormatMultiImage
from dxtbx.model import ScanFactory
from dxtbx.model.beam import Probe

logger = logging.getLogger("dials")


class FormatMRC(Format):
    @staticmethod
    def understand(image_file):
        with FormatMRC.open_file(image_file, "rb") as fh:
            start = fh.read(MAP_ID_OFFSET_BYTES + len(MAP_ID))
            if start[-len(MAP_ID) :] == MAP_ID:
                return True
        return False

    def _start(self):
        """Open the MRC file, read the metadata into an internal dictionary
        self._header_dictionary, add FEI extended metadata if available"""

        with mrcfile.open(self._image_file, header_only=True) as mrc:
            h = mrc.header
            xh = mrc.extended_header

        self._header_dictionary = self._unpack_header(h)
        if len(xh) > 0:
            self._extend_header(xh)

    @staticmethod
    def _unpack_header(header):
        hd = {}
        # What do we need from the header?
        fields = ("nx", "ny", "nz", "mx", "my", "mz")
        for key in fields:
            hd[key] = int(header[key])
        hd["exttyp"] = header["exttyp"].item()

        # For image stacks, NX==MX etc. should always be true. Assert this
        # to ensure we fail on an MRC file of the wrong type.
        assert hd["nx"] == hd["mx"]
        assert hd["ny"] == hd["my"]
        assert hd["nz"] == hd["mz"]

        return hd

    def _extend_header(self, xh):

        # Extensions from FEI headers only
        if not self._header_dictionary["exttyp"].startswith(b"FEI"):
            return

        self._header_dictionary["alphaTilt"] = xh["Alpha tilt"][0]
        self._header_dictionary["integrationTime"] = xh["Integration time"][0]
        self._header_dictionary["tilt_axis"] = xh["Tilt axis angle"][0]
        self._header_dictionary["pixelSpacing"] = xh["Pixel size X"][0]
        assert self._header_dictionary["pixelSpacing"] == xh["Pixel size Y"][0]
        self._header_dictionary["acceleratingVoltage"] = xh["HT"][0]
        self._header_dictionary["camera"] = xh["Camera name"][0]
        self._header_dictionary["binning"] = xh["Binning Width"][0]
        self._header_dictionary["noiseReduction"] = xh["Ceta noise reduction"][0]
        if b"Ceta" in self._header_dictionary["camera"]:
            # Does this ever differ from the Binning Width from the header?
            assert (
                self._header_dictionary["binning"]
                == 4096 / self._header_dictionary["nx"]
            )

        # Make an assumption for Thermo Fisher (Ceta-D and Falcon) detectors.
        # Is this stored anywhere else?
        self._header_dictionary["physicalPixel"] = 14e-6

        def cal_wavelength(V0):
            h = 6.626e-34  # Js, Planck's constant
            m = 9.109e-31  # kg, electron mass
            e = 1.6021766208e-19  # C, electron charge
            c = 3e8  # m/s^2, speed

            # Default to e- wavelength at 200 keV if voltage set to zero
            if V0 == 0:
                V0 = 200000
            return (
                h / sqrt(2 * m * e * V0 * (1 + e * V0 / (2 * m * c * c))) * 1e10
            )  # return wavelength in Angstrom

        self._header_dictionary["wavelength"] = cal_wavelength(
            self._header_dictionary["acceleratingVoltage"]
        )
        self._header_dictionary["cameraLength"] = (
            self._header_dictionary["physicalPixel"]
            * self._header_dictionary["binning"]
        ) / (
            self._header_dictionary["pixelSpacing"]
            * self._header_dictionary["wavelength"]
            * 1e-10
        )

        # Include FEI2 items. This should be detectable by xh["Metadata version"] >= 2
        # but in practice this is not used in all files I have seen. We make the
        # assumption that the first value of each array is relevant for the dataset
        # whether it is an image or a stack
        if self._header_dictionary["exttyp"] != b"FEI2":
            return
        self._header_dictionary["scanRotation"] = xh["Scan rotation"][0]
        self._header_dictionary["diffractionPatternRotation"] = xh[
            "Diffraction pattern rotation"
        ][0]
        self._header_dictionary["imageRotation"] = xh["Image rotation"][0]
        self._header_dictionary["scanModeEnum"] = xh["Scan mode enumeration"][0]
        self._header_dictionary["acquisitionTimeStamp"] = xh["Acquisition time stamp"][
            0
        ]
        self._header_dictionary["detectorCommercialName"] = xh[
            "Detector commercial name"
        ][0]
        self._header_dictionary["startTiltAngle"] = xh["Start tilt angle"][0]
        self._header_dictionary["endTiltAngle"] = xh["End tilt angle"][0]

        self._header_dictionary["tiltPerImage"] = xh["Tilt per image"][0]
        self._header_dictionary["tiltSpeed"] = xh["Tilt speed"][0]
        self._header_dictionary["beamCentreXpx"] = xh["Beam center X pixel"][0]
        self._header_dictionary["beamCentreYpx"] = xh["Beam center Y pixel"][0]
        self._header_dictionary["cfegFlashTimestamp"] = xh["CFEG flash timestamp"][0]
        self._header_dictionary["phasePlatePositionIndex"] = xh[
            "Phase plate position index"
        ][0]
        self._header_dictionary["objectiveApertureName"] = xh[
            "Objective aperture name"
        ][0]

        return

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer."""

        direction = matrix.col((0.0, -1.0, 0.0))
        rot_by_deg = self._header_dictionary.get("tilt_axis", 0.0)
        direction.rotate_around_origin(
            axis=matrix.col((0.0, 0.0, 1.0)), angle=rot_by_deg, deg=True
        )

        return self._goniometer_factory.known_axis(direction)

    def _detector(self):
        """Dummy detector"""

        image_size = (self._header_dictionary["nx"], self._header_dictionary["ny"])

        # Get pixel size, defaulting to 14 um for the Ceta if unknown
        physical_pixel = self._header_dictionary.get("physicalPixel", 1.4e-5)
        binning = self._header_dictionary.get("binning")
        if binning is None:
            if image_size == (2048, 2048):
                binning = 2.0
            else:
                binning = 1.0
        pixel_size = physical_pixel * 1000.0 * binning

        # The best distance measure is calculated from the calibrated pixel
        # size. If this is not available then default to the nominal camera
        # length, or finally to 2.0m
        distance = self._header_dictionary.get("cameraLength", 2.0) * 1000
        try:
            calibrated_pixel_size = self._header_dictionary["pixelSpacing"]  # 1/m
            wavelength = self._header_dictionary["wavelength"] * 1e-10  # m
            distance = pixel_size / (calibrated_pixel_size * wavelength)  # mm
        except KeyError:
            pass

        # Get detector-specific details for TF detectors as discussed with
        # Lingbo Yu. Ceta has gain of > 26 and Ceta and Falcon III both saturate
        # at about 8000.0 for binning=1
        camera = self._header_dictionary.get("camera", b"").lower()
        if b"ceta" in camera:
            gain = 26.0
            saturation = 8000 * binning**2
        elif b"falcon" in camera:
            gain = 1.0
            saturation = 8000 * binning**2
        else:
            gain = 1.0
            saturation = 1e6
        trusted_range = (-1000, saturation)

        # Beam centre not in the header - set to the image centre
        beam_centre = [(pixel_size * i) / 2 for i in image_size]
        detector = self._detector_factory.simple(
            "PAD",
            distance,
            beam_centre,
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            trusted_range,
        )

        for panel in detector:
            panel.set_gain(gain)
        return detector

    def _beam(self):
        """Unpolarized beam model"""

        # Default to 200 keV
        wavelength = self._header_dictionary.get("wavelength", 0.02508)
        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )


class FormatMRCimages(FormatMRC):
    @staticmethod
    def understand(image_file):
        with mrcfile.open(image_file, header_only=True) as mrc:
            return int(mrc.header["nz"]) == 1

    def __init__(self, image_file, **kwargs):
        from dxtbx import IncorrectFormatError

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        Format.__init__(self, image_file, **kwargs)

    def _scan(self):
        """Scan model for this image, filling out any unavailable items with
        dummy values"""

        alpha = self._header_dictionary.get("alphaTilt", 0.0)
        dalpha = self._header_dictionary.get("tiltPerImage", 1.0)
        exposure = self._header_dictionary.get("integrationTime", 0.0)
        oscillation = (alpha, dalpha)
        fname = os.path.split(self._image_file)[-1]
        # assume that the final number before the extension is the image number
        s = fname.split("_")[-1].split(".")[0]
        try:
            index = int(re.match(".*?([0-9]+)$", s).group(1))
        except AttributeError:
            index = 1
        return ScanFactory.make_scan((index, index), exposure, oscillation, {index: 0})

    def get_raw_data(self):

        with mrcfile.open(self._image_file) as mrc:
            image = flex.double(mrc.data.astype("double"))

        return image


class FormatMRCstack(FormatMultiImage, FormatMRC):
    @staticmethod
    def understand(image_file):
        with mrcfile.open(image_file, header_only=True) as mrc:
            return int(mrc.header["nz"]) > 1

    def __init__(self, image_file, **kwargs):
        from dxtbx import IncorrectFormatError

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        FormatMultiImage.__init__(self, **kwargs)
        Format.__init__(self, image_file, **kwargs)

    def get_num_images(self):
        return self._header_dictionary["nz"]

    def get_goniometer(self, index=None):
        return Format.get_goniometer(self)

    def get_detector(self, index=None):
        return Format.get_detector(self)

    def get_beam(self, index=None):
        return Format.get_beam(self)

    def get_scan(self, index=None):
        if index is None:
            return Format.get_scan(self)
        else:
            scan = Format.get_scan(self)
            return scan[index]

    def get_image_file(self, index=None):
        return Format.get_image_file(self)

    def get_raw_data(self, index):

        # Note MRC files use z, y, x ordering
        with mrcfile.mmap(self._image_file) as mrc:
            image = flex.double(mrc.data[index, ...].astype("double"))

        return image

    def _scan(self):
        """Scan model for this stack, filling out any unavailable items with
        dummy values"""

        alpha = self._header_dictionary.get("alphaTilt", 0.0)
        dalpha = self._header_dictionary.get("tiltPerImage", 1.0)
        exposure = self._header_dictionary.get("integrationTime", 0.0)

        oscillation = (alpha, dalpha)
        nframes = self.get_num_images()
        image_range = (1, nframes)
        epochs = [0] * nframes

        return self._scan_factory.make_scan(
            image_range, exposure, oscillation, epochs, deg=True
        )



from __future__ import annotations

import calendar
import os

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

n_pixels_vertical_gaps = 195 * 7 * 4 * 24


@pytest.mark.parametrize(
    "timestamp,multi_panel,masked_pixel_count",
    (
        (None, True, 3053),
        (None, False, 3053 + n_pixels_vertical_gaps),
        (calendar.timegm((2019, 9, 3, 0, 0, 1)), True, 192957),
        (
            calendar.timegm((2019, 9, 3, 0, 0, 1)),
            False,
            192957 + n_pixels_vertical_gaps,
        ),
        (calendar.timegm((2019, 11, 26, 0, 0, 1)), True, 192948),
        (
            calendar.timegm((2019, 11, 26, 0, 0, 1)),
            False,
            192948 + n_pixels_vertical_gaps,
        ),
        (calendar.timegm((2020, 2, 21, 0, 0, 1)), True, 293699),
        (
            calendar.timegm((2020, 2, 21, 0, 0, 1)),
            False,
            293699 + n_pixels_vertical_gaps,
        ),
        (calendar.timegm((2020, 9, 8, 0, 0, 1)), True, 3053),
        (calendar.timegm((2020, 9, 8, 0, 0, 1)), False, 3053 + n_pixels_vertical_gaps),
        (calendar.timegm((2022, 1, 24, 0, 0, 1)), True, 98006),
        (
            calendar.timegm((2022, 1, 24, 0, 0, 1)),
            False,
            98006 + n_pixels_vertical_gaps,
        ),
        (calendar.timegm((2022, 7, 1, 0, 0, 1)), True, 3053),
        (
            calendar.timegm((2022, 7, 1, 0, 0, 1)),
            False,
            3053 + n_pixels_vertical_gaps,
        ),
    ),
)
def test_bad_pixel_mask(
    timestamp, multi_panel, masked_pixel_count, dials_regression, mocker
):
    if timestamp is not None:
        # Fool the format class into masking out bad modules
        mocked_timestamp = mocker.patch(
            "dxtbx.format.FormatCBFMiniPilatusDLS12M.get_pilatus_timestamp"
        )
        mocked_timestamp.return_value = timestamp

    image_path = os.path.join(
        dials_regression, "image_examples", "DLS_I23", "germ_13KeV_0001.cbf"
    )
    experiments = ExperimentListFactory.from_filenames(
        [image_path], format_kwargs={"multi_panel": multi_panel}
    )
    assert (
        sum(mask.count(False) for mask in experiments[0].imageset.get_mask(0))
        == masked_pixel_count
    )



from __future__ import annotations

import math

import numpy as np


def rotate_and_average(data, angle, deg=False, mask=None):
    """
    Angle is the rotation angle in the coordinate system of x along fast, y along slow, such that angle increases as it rotates around z, where z completes a right-handed system.
    """
    if deg:
        angle *= math.pi / 180
    angle = -angle

    ny, nx = np.shape(data)

    xx, yy = np.meshgrid(np.arange(nx), np.arange(ny))
    xx_yy = np.row_stack((xx.ravel(), yy.ravel()))
    R = np.array(((np.cos(angle), -np.sin(angle)), (np.sin(angle), np.cos(angle))))
    xx_yy_rotated = np.matmul(R, xx_yy)
    xx_rotated = xx_yy_rotated[0, :].reshape((ny, nx))
    # yy_rotated = xx_yy_rotated[1, :].reshape((ny, nx))
    bin_x = np.arange(-nx, nx)
    bin_x_centers = (bin_x[1:] + bin_x[:-1]) / 2

    if mask is None:
        mask = np.full(data.shape, True)

    counts_rotated, _ = np.histogram(xx_rotated[mask], bins=bin_x)
    summed_rotated, _ = np.histogram(xx_rotated[mask], bins=bin_x, weights=data[mask])
    sel = counts_rotated > 0
    average_rotated = summed_rotated[sel] / counts_rotated[sel]

    return bin_x_centers[sel], average_rotated



from __future__ import annotations

import functools
import sys
import time

import numpy as np
import serialtbx.detector.xtc
import serialtbx.util

from libtbx.phil import parse
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format, abstract
from dxtbx.format.FormatMultiImage import FormatMultiImage, Reader
from dxtbx.format.FormatStill import FormatStill
from dxtbx.model import Spectrum
from dxtbx.util.rotate_and_average import rotate_and_average

try:
    import psana
except ImportError:
    psana = None
except TypeError:
    # Check if SIT_* environment variables are set
    import os

    if os.environ.get("SIT_ROOT"):
        # Variables are present, so must have been another error
        raise
    psana = None

locator_str = """
  experiment = None
    .type = str
    .help = Experiment identifier, e.g. mfxo1916
  run = None
    .type = ints
    .help = Run number or a list of runs to process
  mode = idx
    .type = str
    .help = Mode for reading the xtc data (see LCLS documentation)
  data_source = None
    .type = str
    .help = Complete LCLS data source.  Overrides experiment and run.  Example: \
            exp=mfxo1916:run=20:smd \
            More info at https://confluence.slac.stanford.edu/display/PSDM/Manual#Manual-Datasetspecification
  detector_address = None
    .type = str
    .multiple = True
    .help = detector used for collecting the data at LCLS
  calib_dir = None
    .type = str
    .help = Specify path to custom calib directory if needed
  use_ffb = False
    .type = bool
    .help = Run on the ffb if possible. Only for active users!
  wavelength_delta_k = 0
    .type = float
    .help = Correction factor, needed during 2014
  wavelength_offset = None
    .type = float
    .help = Optional constant shift to apply to each wavelength
  spectrum_address = FEE-SPEC0
    .type = str
    .help = Address for incident beam spectrometer
  spectrum_eV_per_pixel = None
    .type = float
    .help = If not None, use the FEE spectrometer to determine the wavelength. \
            spectrum_eV_offset should be also specified. A weighted average of \
            the horizontal projection of the per-shot FEE spectrometer is used.\
            The equation for each pixel eV is \
            eV = (spectrum_eV_per_pixel * pixel_number) + spectrum_eV_offset
  spectrum_eV_offset = None
    .type = float
    .help = See spectrum_eV_per_pixel
  spectrum_rotation_angle = None
    .type = float
    .help = If present, first rotate the spectrum image a given amount in \
            degrees. Only applies to 2D spectrometers
  spectrum_pedestal = None
    .type = path
    .help = Path to pickled pedestal file to subtract from the pedestal
  filter {
    evr_address = evr1
      .type = str
      .help = Address for evr object which stores event codes. Should be evr0,\
              evr1, or evr2.
    required_present_codes = None
      .type = int
      .multiple = True
      .help = These codes must be present to keep the event
    required_absent_codes = None
      .type = int
      .multiple = True
      .help = These codes must be absent to keep the event
    pre_filter = False
      .type = bool
      .help = If True, read the event codes for all events up front, and \
              apply the filter then. Otherwise, apply when loading an \
              event.
  }
"""
locator_scope = parse(locator_str)


class XtcReader(Reader):
    def nullify_format_instance(self):
        """No-op for XTC streams. No issue with multiprocessing."""


@abstract
class FormatXTC(FormatMultiImage, FormatStill, Format):
    def __init__(self, image_file, **kwargs):

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self.lazy = kwargs.get("lazy", True)
        FormatMultiImage.__init__(self, **kwargs)
        FormatStill.__init__(self, image_file, **kwargs)
        Format.__init__(self, image_file, **kwargs)
        self.current_index = None
        self.current_event = None
        self._psana_runs = {}  # empty container, to prevent breaking other formats
        if "locator_scope" in kwargs:
            self.params = FormatXTC.params_from_phil(
                master_phil=kwargs["locator_scope"], user_phil=image_file, strict=True
            )
        else:
            self.params = FormatXTC.params_from_phil(
                master_phil=locator_scope, user_phil=image_file, strict=True
            )
        assert self.params.mode in [
            "idx",
            "smd",
        ], "idx or smd mode should be used for analysis (idx is often faster)"

        self._ds = FormatXTC._get_datasource(image_file, self.params)
        self._evr = None
        self.populate_events()

        self._cached_psana_detectors = {}
        self._beam_index = None
        self._beam_cache = None
        self._initialized = True
        self._fee = None

        if self.params.spectrum_pedestal:
            from libtbx import easy_pickle

            self._spectrum_pedestal = easy_pickle.load(self.params.spectrum_pedestal)
        else:
            self._spectrum_pedestal = None

    @staticmethod
    def understand(image_file):
        """Extracts the datasource and detector_address from the image_file and then feeds it to PSANA
        If PSANA fails to read it, then input may not be an xtc/smd file. If success, then OK.
        If detector_address is not provided, a command line promp will try to get the address
        from the user"""
        if not psana:
            return False
        try:
            params = FormatXTC.params_from_phil(locator_scope, image_file)
        except Exception:
            return False
        if params is None:
            return False

        try:
            FormatXTC._get_datasource(image_file, params)
        except Exception:
            return False
        return True

    @staticmethod
    def params_from_phil(master_phil, user_phil, strict=False):
        """Read the locator file"""
        try:
            user_input = parse(file_name=user_phil)
            working_phil, unused = master_phil.fetch(
                sources=[user_input], track_unused_definitions=True
            )
            unused_args = ["%s=%s" % (u.path, u.object.words[0].value) for u in unused]
            if len(unused_args) > 0 and strict:
                for unused_arg in unused_args:
                    print(unused_arg)
                print(
                    "Incorrect of unused parameter in locator file. Please check and retry"
                )
                return None
            params = working_phil.extract()
            return params
        except Exception:
            return None

    @classmethod
    def get_reader(cls):
        """
        Return a reader class
        """
        return functools.partial(XtcReader, cls)

    def populate_events(self):
        """Read the timestamps from the XTC stream.  Assumes the psana idx mode of reading data.
        Handles multiple LCLS runs by concatenating the timestamps from multiple runs together
        in a single list and creating a mapping."""
        if self.params.mode == "idx":
            if hasattr(self, "times") and len(self.times) > 0:
                return
        elif self.params.mode == "smd":
            if hasattr(self, "run_mapping") and self.run_mapping:
                return

        if not self._psana_runs:
            self._psana_runs = self._get_psana_runs(self._ds)

        self.times = []
        self.run_mapping = {}

        if self.params.mode == "idx":
            for run in self._psana_runs.values():
                times = run.times()
                if (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                ) and self.params.filter.pre_filter:
                    times = [t for t in times if self.filter_event(run.event(t))]
                self.run_mapping[run.run()] = (
                    len(self.times),
                    len(self.times) + len(times),
                    run,
                )
                self.times.extend(times)
            self.n_images = len(self.times)

        elif self.params.mode == "smd":
            self._ds = FormatXTC._get_datasource(self._image_file, self.params)
            for event in self._ds.events():
                run = event.run()
                if run not in self.run_mapping:
                    self.run_mapping[run] = []
                if (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                ) and self.params.filter.pre_filter:
                    if self.filter_event(event):
                        self.run_mapping[run].append(event)
                else:
                    self.run_mapping[run].append(event)
            total = 0
            remade_mapping = {}
            for run in sorted(self.run_mapping):
                start = total
                end = len(self.run_mapping[run]) + total
                total += len(self.run_mapping[run])
                events = self.run_mapping[run]
                remade_mapping[run] = start, end, run, events
            self.run_mapping = remade_mapping
            self.n_images = sum(
                [
                    self.run_mapping[r][1] - self.run_mapping[r][0]
                    for r in self.run_mapping
                ]
            )

    def filter_event(self, evt):
        """Return True to keep the event, False to reject it."""
        if not (
            self.params.filter.required_present_codes
            or self.params.filter.required_absent_codes
        ):
            return True
        if not self._evr:
            self._evr = psana.Detector(self.params.filter.evr_address)
        codes = self._evr.eventCodes(evt)

        if self.params.filter.required_present_codes and not all(
            [c in codes for c in self.params.filter.required_present_codes]
        ):
            return False
        if self.params.filter.required_absent_codes and any(
            [c in codes for c in self.params.filter.required_absent_codes]
        ):
            return False
        return True

    def get_run_from_index(self, index=None):
        """Look up the run number given an index"""
        if index is None:
            index = 0
        for run_number in self.run_mapping:
            start, stop, run = self.run_mapping[run_number][0:3]
            if index >= start and index < stop:
                return run
        raise IndexError("Index is not within bounds")

    def _get_event(self, index):
        """Retrieve a psana event given and index. This is the slow step for reading XTC streams,
        so implement a cache for the last read event."""
        if index == self.current_index:
            return self.current_event
        else:
            self.current_index = index
            if self.params.mode == "idx":
                evt = self.get_run_from_index(index).event(self.times[index])
            elif self.params.mode == "smd":
                for run_number in self.run_mapping:
                    start, stop, run, events = self.run_mapping[run_number]
                    if index >= start and index < stop:
                        evt = events[index - start]
            if (
                (
                    self.params.filter.required_present_codes
                    or self.params.filter.required_absent_codes
                )
                and not self.params.filter.pre_filter
                and not self.filter_event(evt)
            ):
                evt = None
            self.current_event = evt
            return self.current_event

    @staticmethod
    def _get_datasource(image_file, params):
        """Construct a psana data source object given the locator parameters"""
        if params.calib_dir is not None:
            psana.setOption("psana.calib-dir", params.calib_dir)
        if params.data_source is None:
            if (
                params.experiment is None
                or params.run is None
                or params.mode is None
                or len(params.run) == 0
            ):
                return False
            img = "exp=%s:run=%s:%s" % (
                params.experiment,
                ",".join(["%d" % r for r in params.run]),
                params.mode,
            )

            if params.use_ffb:
                # as ffb is only at SLAC, ok to hardcode /reg/d here
                img += ":dir=/reg/d/ffb/%s/%s/xtc" % (
                    params.experiment[0:3],
                    params.experiment,
                )
        else:
            img = params.data_source
        return psana.DataSource(img)

    @staticmethod
    def _get_psana_runs(datasource):
        """
        Extracts the runs,
        These can only be extracted once,
        only call this method after datasource is set
        """
        # this is key,value = run_integer, psana.Run, e.g. {62: <psana.Run(@0x7fbd0e23c990)>}
        psana_runs = {r.run(): r for r in datasource.runs()}
        return psana_runs

    def _get_psana_detector(self, run):
        """Returns the psana detector for the given run"""
        if run.run() not in self._cached_psana_detectors:
            assert len(self.params.detector_address) == 1
            self._cached_psana_detectors[run.run()] = psana.Detector(
                self.params.detector_address[0], run.env()
            )
        return self._cached_psana_detectors[run.run()]

    def get_psana_timestamp(self, index):
        """Get the cctbx.xfel style event timestamp given an index"""
        evt = self._get_event(index)
        if not evt:
            return None
        time = evt.get(psana.EventId).time()
        # fid = evt.get(psana.EventId).fiducials()

        sec = time[0]
        nsec = time[1]

        return serialtbx.util.timestamp((sec, nsec / 1e6))

    def get_num_images(self):
        return self.n_images

    def get_beam(self, index=None):
        return self._beam(index)

    def _beam(self, index=None):
        """Returns a simple model for the beam"""
        if index is None:
            index = 0
        if self._beam_index != index:
            self._beam_index = index
            evt = self._get_event(index)
            if not evt:
                self._beam_cache = None
                return None
            spectrum = self.get_spectrum(index)
            if spectrum:
                wavelength = spectrum.get_weighted_wavelength()
            else:
                wavelength = serialtbx.detector.xtc.evt_wavelength(
                    evt, delta_k=self.params.wavelength_delta_k
                )
            if wavelength is None:
                self._beam_cache = None
            else:
                if self.params.wavelength_offset is not None:
                    wavelength += self.params.wavelength_offset
                self._beam_cache = self._beam_factory.simple(wavelength)
            s, nsec = evt.get(psana.EventId).time()
            evttime = time.gmtime(s)
            if (
                evttime.tm_year == 2020 and evttime.tm_mon >= 7
            ) or evttime.tm_year > 2020:
                if self._beam_cache is not None:
                    self._beam_cache.set_polarization_normal((1, 0, 0))

        return self._beam_cache

    def get_spectrum(self, index=None):
        if index is None:
            index = 0
        if self.params.spectrum_eV_per_pixel is None:
            return None

        evt = self._get_event(index)
        if not evt:
            return None
        if self._fee is None:
            self._fee = psana.Detector(self.params.spectrum_address)
        if self._fee is None:
            return None
        try:
            fee = self._fee.get(evt)
            y = fee.hproj()
            if self.params.spectrum_pedestal:
                y = y - self._spectrum_pedestal.as_numpy_array()

        except AttributeError:  # Handle older spectometers without the hproj method
            try:
                img = self._fee.image(evt)
            except AttributeError:
                return None
            if self.params.spectrum_rotation_angle is None:
                x = (
                    self.params.spectrum_eV_per_pixel * np.array(range(img.shape[1]))
                ) + self.params.spectrum_eV_offset
                y = img.mean(axis=0)  # Collapse 2D image to 1D trace
            else:
                mask = img == 2**16 - 1
                mask = np.invert(mask)

                x, y = rotate_and_average(
                    img, self.params.spectrum_rotation_angle, deg=True, mask=mask
                )
                x = (
                    self.params.spectrum_eV_per_pixel * x
                ) + self.params.spectrum_eV_offset

        else:
            x = (
                self.params.spectrum_eV_per_pixel * np.array(range(len(y)))
            ) + self.params.spectrum_eV_offset
        return Spectrum(flex.double(x), flex.double(y))

    def get_goniometer(self, index=None):
        return None

    def get_scan(self, index=None):
        return None


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXTC.understand(arg))



"""
Experimental format for Gatan Digital Micrograph DM4 files. See
https://personal.ntu.edu.sg/cbb/info/dmformat/index.html
"""

from __future__ import annotations

import struct

from boost_adaptbx.boost.python import streambuf
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.ext import (
    read_float32,
    read_int16,
    read_int32,
    read_uint8,
    read_uint16,
    read_uint32,
)
from dxtbx.format.Format import Format
from dxtbx.format.FormatMultiImage import FormatMultiImage
from dxtbx.model.beam import Probe


def read_tag(f, byteorder):
    """read a tag from an open and correctly positioned DM4 file, omitting tag
    data for groups and arrays (to be fulfilled later as needed)"""

    tag = {}
    tag_type = struct.unpack("B", f.read(1))[0]
    if tag_type == 0:  # EOF
        return None
    elif tag_type == 20:
        return read_tag_directory(f, byteorder)
    elif tag_type != 21:
        # unrecognized format
        raise ValueError("Unrecognised tag type while reading data")

    ltname = struct.unpack(">H", f.read(2))[0]
    try:
        tag["tname"] = f.read(ltname).decode()
    except UnicodeDecodeError:
        tag["tname"] = ""

    tag["tlen"] = struct.unpack(">Q", f.read(8))[0]
    tag["tend"] = f.tell() + tag["tlen"]
    assert f.read(4) == b"%%%%"
    tag["ninfo"] = struct.unpack(">Q", f.read(8))[0]
    tag["info"] = struct.unpack(">" + "Q" * tag["ninfo"], f.read(8 * tag["ninfo"]))

    if tag["ninfo"] == 1:  # single entry tag
        lookup = FormatGatanDM4._tag_data_type.get(tag["info"][0])

        if lookup is None:  # unknown tag data type - skip this tag
            tag["value"] = None
            f.seek(tag["tend"])
            return tag

        # read the tag value
        data_type, nbytes = lookup
        tag["value"] = struct.unpack(byteorder + data_type, f.read(nbytes))[0]

    else:  # group or array, don't read now, just record the file pos
        tag["offset"] = f.tell()
        f.seek(tag["tend"])

    return tag


def read_tag_directory(f, byteorder):
    """read a tag directory from an open file. The first byte that identifies
    the tag directory is expected to have been read already"""

    tag_dir = {}

    ltname = struct.unpack(">H", f.read(2))[0]
    tag_dir["tname"] = f.read(ltname).decode()
    tag_dir["tlen"] = struct.unpack(">Q", f.read(8))[0]
    tag_dir["sortf"] = struct.unpack("B", f.read(1))[0]
    tag_dir["closef"] = struct.unpack("B", f.read(1))[0]
    tag_dir["ntags"] = struct.unpack(">Q", f.read(8))[0]
    tag_dir["tags"] = []

    for i in range(tag_dir["ntags"]):
        tag = read_tag(f, byteorder)
        if tag is None:
            break  # Give up if EOF reached
        tag_dir["tags"].append(tag)

    return tag_dir


def print_tag_hierarchy(tag_list, prefix=""):
    for tag in tag_list:
        name = tag["tname"]
        if name == "":
            name = "NONAME"
        if "tags" in tag:  # tag directory
            print(prefix + f"DIR: {name}")
            prefix += "  "
            print_tag_hierarchy(tag["tags"], prefix)
            prefix = prefix[:-2]
        else:  # standard tag
            val = tag.get("value")
            if val is not None:
                val = ": " + str(val)
            else:
                val = ": " + str(tag["info"])
            print(prefix + name + val)


def search_tag_hierarchy(name, tag_list):
    """Search a list of tags for any with the given name."""
    result = []
    for tag in tag_list:
        if tag["tname"] == name:
            result.extend([tag])
        else:
            if "tags" in tag:
                result.extend(search_tag_hierarchy(name, tag["tags"]))
    return result


def extract_image_data(image_file, offset, ntags, byteord):
    # read tags
    with FormatGatanDM4.open_file(image_file, "rb") as f:

        f.seek(offset)

        root = []
        for i in range(ntags):
            root.append(read_tag(f, byteord))

        # print_tag_hierarchy(root)

    # Get indices of images to ignore because they are thumbnails
    ignore_idx = []
    for tag in root:
        if tag["tname"] == "Thumbnails":
            vals = [v["value"] for v in search_tag_hierarchy("ImageIndex", [tag])]
            ignore_idx.extend(vals)

    # Get the ImageList and remove thumbnails
    imlist = []
    for tag in root:
        if tag["tname"] == "ImageList":
            data = [t for i, t in enumerate(tag["tags"]) if i not in ignore_idx]
            imlist.extend(data)

    # Should only be one item left now! Extract ImageData (leaving ImageTags,
    # Name and UniqueID aside)
    assert len(imlist) == 1
    image_data = search_tag_hierarchy("ImageData", imlist)
    assert len(image_data) == 1
    return image_data


def extract_num_images(image_file):
    """Extract the number of images contained in the file. The format is horrible
    to work with and this operation has many steps"""

    header = FormatGatanDM4._read_header(image_file)
    byteord = "<" if header["byteord"] == 1 else ">"

    image_data = extract_image_data(
        image_file, offset=header["offset"], ntags=header["ntags"], byteord=byteord
    )

    dimensions = search_tag_hierarchy("Dimensions", image_data)[0]["tags"]
    image_size = dimensions[0]["value"], dimensions[1]["value"]

    # extract the number of images
    data = search_tag_hierarchy("Data", image_data)
    assert len(data) == 1
    data = data[0]
    image_num_elements = image_size[0] * image_size[1]
    total_num_elements = data["info"][2]
    num_images = total_num_elements // image_num_elements

    return num_images


class FormatGatanDM4(Format):
    """An image reading class for images in Gatan Digital Micrograph DM4 format.
    Concrete subclasses handle either one-file-per image, or all images in a
    single stack.

    The header does not contain useful information about the geometry, therefore
    we will construct dummy objects and expect to override on import using
    site.phil."""

    # set up tag data types and lengths
    _tag_data_type = {
        2: ("h", 2),
        3: ("l", 4),
        4: ("H", 2),
        5: ("L", 4),
        6: ("f", 4),
        7: ("d", 8),
        8: ("?", 1),
        9: ("c", 1),
        10: ("b", 1),
        11: ("q", 8),
        15: ("struct", None),
        18: ("string", None),
        20: ("array", None),
    }

    # set up image data types and lengths
    _image_data_type = {
        0: ("null", None),
        1: ("h", 2),
        2: ("f", 4),
        3: ("complex", 8),
        4: ("obsolete", None),
        5: ("complex", 4),
        6: ("B", 1),
        7: ("l", 4),
        8: ("_rgb", 4),
        9: ("b", 1),
        10: ("H", 2),
        11: ("I", 4),
        12: ("d", 8),
        13: ("complex", 16),
        14: ("?",),
        23: ("rgba", 4),
    }

    @staticmethod
    def understand(image_file):

        try:
            header = FormatGatanDM4._read_header(image_file)
        except struct.error:
            return False

        if header["version"] != 4:
            return False
        if header["byteord"] not in [0, 1]:
            return False
        if header["sortf"] not in [0, 1]:
            return False
        if header["closef"] not in [0, 1]:
            return False

        return True

    @staticmethod
    def _read_header(image_file):
        """read the minimal header metadata and structure of the root tag
        directory"""

        hd = {}

        with FormatGatanDM4.open_file(image_file, "rb") as f:

            hd["version"] = struct.unpack(">I", f.read(4))[0]

            if hd["version"] == 3:
                hd["rootlen"] = struct.unpack(">I", f.read(4))[0]
            else:
                hd["rootlen"] = struct.unpack(">Q", f.read(8))[0]

            hd["byteord"] = struct.unpack(">I", f.read(4))[0]

            hd["sortf"] = struct.unpack("B", f.read(1))[0]

            hd["closef"] = struct.unpack("B", f.read(1))[0]

            if hd["version"] == 3:
                hd["ntags"] = struct.unpack(">I", f.read(4))[0]
            else:
                hd["ntags"] = struct.unpack(">Q", f.read(8))[0]

            hd["offset"] = f.tell()

        return hd

    def _start(self):
        """Open the image file, read useful metadata into an internal dictionary
        and read the tags to discover where the image data are and how to read it"""

        self._header_dictionary = self._read_header(self._image_file)

        self._byteord = "<" if self._header_dictionary["byteord"] == 1 else ">"

        image_data = extract_image_data(
            self._image_file,
            offset=self._header_dictionary["offset"],
            ntags=self._header_dictionary["ntags"],
            byteord=self._byteord,
        )

        # data type and size
        lookup = search_tag_hierarchy("DataType", image_data)[0]["value"]
        lookup = FormatGatanDM4._image_data_type.get(lookup)
        if lookup is None:
            raise ValueError("Unrecognised data type for image data")
        else:
            self._data_type, self._data_size = lookup
        if self._data_type not in ["h", "f", "B", "l", "b", "H", "I", "d", "?"]:
            raise ValueError("Unrecognised data type for image data")

        # image dimensions
        dimensions = search_tag_hierarchy("Dimensions", image_data)[0]["tags"]
        self._image_size = dimensions[0]["value"], dimensions[1]["value"]

        # work out how to read the data
        data = search_tag_hierarchy("Data", image_data)
        assert len(data) == 1
        data = data[0]
        self._image_num_elements = self._image_size[0] * self._image_size[1]
        self._total_num_elements = data["info"][2]
        self._num_images = self._total_num_elements // self._image_num_elements
        assert self._total_num_elements % self._image_num_elements == 0
        self._data_offset = data["offset"]

        return

    def _goniometer(self):
        """Dummy goniometer, 'vertical' as the images are viewed"""

        return self._goniometer_factory.known_axis((0, -1, 0))

    def _detector(self):
        """Dummy detector. It may be possible to extract some of the required
        details from the image metadata, but not sure how consistent they are."""

        # Assuming Gatan OneView camera, with 15um pixels, but trusted range is
        # (probably) wrong.
        pixel_size = 0.015, 0.015
        image_size = self._image_size
        trusted_range = (0, 65535)
        distance = 2000
        beam_centre = [(p * i) / 2 for p, i in zip(pixel_size, image_size)]
        d = self._detector_factory.simple(
            "PAD",
            distance,
            beam_centre,
            "+x",
            "-y",
            pixel_size,
            image_size,
            trusted_range,
        )
        # Not sure what the gain is. A Gatan OneView has a scintillator and fibre
        # optic, so is not a direct detection camera and probably has gain != 1
        # for p in d: p.set_gain(8)
        return d

    def _beam(self):
        """Dummy unpolarized beam, energy 200 keV"""

        wavelength = 0.02508
        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )

    def _scan(self):
        """Dummy scan"""

        image_range = (1, self._num_images)
        exposure_times = 0.0
        oscillation = (0, 0.5)
        epochs = [0] * self._num_images

        return self._scan_factory.make_scan(
            image_range, exposure_times, oscillation, epochs, deg=True
        )

    def _read_raw_data(self, f):
        """Read raw data from a positioned file"""

        # is this image a type we can read?
        assert self._data_type in ["h", "f", "B", "l", "b", "H", "I", "d"]

        if self._data_type == "f":
            raw_data = read_float32(streambuf(f), self._image_num_elements)
        elif self._data_type == "B":
            raw_data = read_uint8(streambuf(f), self._image_num_elements)
        elif self._data_type == "h":
            raw_data = read_int16(streambuf(f), self._image_num_elements)
        elif self._data_type == "H":
            raw_data = read_uint16(streambuf(f), self._image_num_elements)
        elif self._data_type == "l":
            raw_data = read_int32(streambuf(f), self._image_num_elements)
        elif self._data_type == "I":
            raw_data = read_uint32(streambuf(f), self._image_num_elements)

        # no C++ reader for remaining types (should be unusual anyway)
        else:
            vals = struct.unpack(
                self._byteord + self._data_type * self._image_num_elements,
                f.read(self._data_size * self._image_num_elements),
            )
            if self._data_type == "d":
                raw_data = flex.double(vals)
            if self._data_type in ["b", "?"]:
                raw_data = flex.int(vals)

        raw_data.reshape(flex.grid(self._image_size[1], self._image_size[0]))
        return raw_data


class FormatGatanDM4Images(FormatGatanDM4):
    @staticmethod
    def understand(image_file):
        return extract_num_images(image_file) == 1

    def __init__(self, image_file, **kwargs):

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        FormatGatanDM4.__init__(self, image_file, **kwargs)

    def get_raw_data(self):

        with FormatGatanDM4.open_file(self._image_file, "rb") as f:
            f.seek(self._data_offset)
            return self._read_raw_data(f)


class FormatGatanDM4Stack(FormatMultiImage, FormatGatanDM4):
    @staticmethod
    def understand(image_file):
        return extract_num_images(image_file) > 1

    def __init__(self, image_file, **kwargs):

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        FormatMultiImage.__init__(self, **kwargs)
        FormatGatanDM4.__init__(self, image_file, **kwargs)

    def get_raw_data(self, index):

        with FormatGatanDM4.open_file(self._image_file, "rb") as f:
            f.seek(self._data_offset)

            skip_bytes = index * self._image_num_elements * self._data_size
            f.seek(skip_bytes, whence=1)

            return self._read_raw_data(f)

    def get_num_images(self):
        return self._num_images

    def get_goniometer(self, index=None):
        return Format.get_goniometer(self)

    def get_detector(self, index=None):
        return Format.get_detector(self)

    def get_beam(self, index=None):
        return Format.get_beam(self)

    def get_scan(self, index=None):
        if index is None:
            return Format.get_scan(self)
        else:
            scan = Format.get_scan(self)
            return scan[index]

    def get_image_file(self, index=None):
        return Format.get_image_file(self)



"""
Format class for the PAL XFEL raw data format
"""

from __future__ import annotations

import sys

import h5py

from scitbx.array_family import flex

from dxtbx.format.FormatHDF5 import FormatHDF5


class FormatHDF5PAL(FormatHDF5):
    @staticmethod
    def understand(image_file):
        with h5py.File(image_file, "r") as h5_handle:
            if len(h5_handle) != 1:
                return False

            for key in h5_handle:
                if not key.startswith("R"):
                    return False
                try:
                    int(key.lstrip("R"))
                except ValueError:
                    return False
                for subkey in h5_handle[key]:
                    if subkey not in ("header", "scan_dat"):
                        return False
                    if subkey == "scan_dat":
                        if "raymx_header" not in h5_handle[key][subkey]:
                            return False
        return True

    def _start(self):
        self._h5_handle = h5py.File(self.get_image_file(), "r")
        self._run = list(self._h5_handle.keys())[0]

        # currently hardcoded to Rayonix MX225HS
        self._detector_size = 225  # mm
        self._max_pixels = 5760
        frame_1 = self.get_raw_data(0)
        assert frame_1.focus()[0] == frame_1.focus()[1]
        self._binning = self._max_pixels // frame_1.focus()[0]

    def get_raw_data(self, index=None):
        if index is None:
            index = 0

        data = self._h5_handle[self._run]["scan_dat/raymx_data"][index]
        # return flex.int(int) # this crashes!
        # return flex.int(data.astype(np.int)) # this doesn't work! (data is read incorrectly)
        return flex.double(data.astype(float))

    def get_num_images(self):
        return len(self._h5_handle[self._run]["scan_dat/N"][()])

    def _detector(self):
        distance = self._h5_handle[self._run]["header/detector_0_distance"][()]
        image_size = self._max_pixels // self._binning
        pixel_size = self._detector_size / image_size
        beam_x = 0.5 * self._detector_size
        beam_y = 0.5 * self._detector_size
        trusted_range = (
            0,
            65534,
        )  # note one less than normal due to how bad pixels are encoded for this detector

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            (image_size, image_size),
            trusted_range,
            [],
        )

    def _beam(self, index=None):
        if index is None:
            index = 0
        return self._beam_factory.simple(
            self._h5_handle[self._run]["scan_dat/photon_wavelength"][index]
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatHDF5PAL.understand(arg))



# LIBTBX_SET_DISPATCHER_NAME dxtbx.image2pickle
# LIBTBX_SET_DISPATCHER_NAME cxi.image2pickle
"""
Convert images of any extant format to pickle files suitable for processing with
cxi.index.  Note, oscillation values are not preserved.
"""

from __future__ import annotations

import math
import os
import sys

import numpy as np
import serialtbx.detector.cspad
import serialtbx.util

import libtbx.option_parser
from libtbx import easy_pickle
from libtbx.utils import Usage
from scitbx.array_family import flex

import dxtbx.util


def crop_image_pickle(
    data, preserve_active_areas_even_though_cropping_would_invalidate_them=False
):
    """
    Given an image pickle dictionary, crop the pixels such that the beam center is as close
    as possile to the image center.  Then adjust SIZE1/SIZE2, ACTIVE_AREAS and the beam
    center accordingly.
    @param data The image dictionary of interest
    """
    # only one active area is allowed, and it should be the size of the image.

    if preserve_active_areas_even_though_cropping_would_invalidate_them is False:
        test = flex.int([0, 0, data["SIZE1"], data["SIZE2"]]) == data["ACTIVE_AREAS"]
        assert test.all_eq(True)

    # retrieve parameters from the dictionary
    pixel_size = data["PIXEL_SIZE"]
    beam_x = int(round(data["BEAM_CENTER_X"] / pixel_size))
    beam_y = int(round(data["BEAM_CENTER_Y"] / pixel_size))
    width = data["SIZE1"]
    height = data["SIZE2"]
    pixels = data["DATA"]

    # the new image will be twice the size as the smallest distance between the
    # beam center and one of the edges of the images
    new_half_size = min([beam_x, width - beam_x, beam_y, height - beam_y])
    new_size = new_half_size * 2
    min_x = beam_x - new_half_size
    min_y = beam_y - new_half_size
    pixels = pixels[min_y : min_y + new_size, min_x : min_x + new_size]
    assert pixels.focus()[0] == pixels.focus()[1]

    # save the results
    data["DATA"] = pixels
    data["SIZE1"] = new_size
    data["SIZE2"] = new_size
    data["BEAM_CENTER_X"] -= min_x * pixel_size
    data["BEAM_CENTER_Y"] -= min_y * pixel_size
    if preserve_active_areas_even_though_cropping_would_invalidate_them is False:
        data["ACTIVE_AREAS"] = flex.int([0, 0, new_size, new_size])

    return data


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    args = args or sys.argv[1:]
    progname = os.getenv("LIBTBX_DISPATCHER_NAME")
    if not progname or progname.endswith(".python"):
        progname = "%prog"
    command_line = (
        libtbx.option_parser.option_parser(
            usage=f"{progname} [-v] [-c] [-s] [-w wavelength] [-d distance] [-p pixel_size] [-x beam_x] [-y beam_y] [-o overload] files"
        )
        .option(
            None,
            "--verbose",
            "-v",
            action="store_true",
            default=False,
            dest="verbose",
            help="Print more information about progress",
        )
        .option(
            None,
            "--crop",
            "-c",
            action="store_true",
            default=False,
            dest="crop",
            help="Crop the image such that the beam center is in the middle",
        )
        .option(
            None,
            "--skip_converted",
            "-s",
            action="store_true",
            default=False,
            dest="skip_converted",
            help="Skip converting if an image already exist that matches the destination file name",
        )
        .option(
            None,
            "--wavelength",
            "-w",
            type="float",
            default=None,
            dest="wavelength",
            help="Override the image's wavelength (angstroms)",
        )
        .option(
            None,
            "--distance",
            "-d",
            type="float",
            default=None,
            dest="distance",
            help="Override the detector distance (mm)",
        )
        .option(
            None,
            "--pixel_size",
            "-p",
            type="float",
            default=None,
            dest="pixel_size",
            help="Override the detector pixel size (mm)",
        )
        .option(
            None,
            "--beam_x",
            "-x",
            type="float",
            default=None,
            dest="beam_center_x",
            help="Override the beam x position (pixels)",
        )
        .option(
            None,
            "--beam_y",
            "-y",
            type="float",
            default=None,
            dest="beam_center_y",
            help="Override the beam y position (pixels)",
        )
        .option(
            None,
            "--overload",
            "-o",
            type="float",
            default=None,
            dest="overload",
            help="Override the detector overload value (ADU)",
        )
    ).process(args)

    paths = command_line.args
    if len(paths) <= 0:
        raise Usage("No files specified")

    for imgpath in paths:
        if command_line.options.verbose:
            print("Reading %s" % (imgpath))

        try:
            img = dxtbx.load(imgpath)
        except OSError:
            img = None

        if img is None:
            try:
                raw_data = np.loadtxt(imgpath)
                raw_data = flex.double(raw_data.astype(np.double))
            except ValueError:
                raise Usage("Couldn't load %s, no supported readers" % imgpath)

            detector = None
            beam = None
            scan = None
            is_multi_image = False
        else:
            try:
                raw_data = img.get_raw_data()
                is_multi_image = False
            except TypeError:
                raw_data = img.get_raw_data(0)
                is_multi_image = True
            detector = img.get_detector()
            beam = img.get_beam()
            scan = img.get_scan()

        if detector is None:
            if command_line.options.distance is None:
                raise Usage("Can't get distance from image. Override with -d")
            if command_line.options.pixel_size is None:
                raise Usage("Can't get pixel size from image. Override with -p")
            if command_line.options.overload is None:
                raise Usage("Can't get overload value from image. Override with -o")
            distance = command_line.options.distance
            pixel_size = command_line.options.pixel_size
            overload = command_line.options.overload
        else:
            detector = detector[0]
            if command_line.options.distance is None:
                distance = detector.get_distance()
            else:
                distance = command_line.options.distance

            if command_line.options.pixel_size is None:
                pixel_size = detector.get_pixel_size()[0]
            else:
                pixel_size = command_line.options.pixel_size

            if command_line.options.overload is None:
                # Here an assumption is made that the overload value is 1 ADU greater
                # than the maximum trusted value of the detector. This is not
                # correct for detectors with floating point raw data.
                overload = detector.get_trusted_range()[1] + 1
            else:
                overload = command_line.options.overload

        if beam is None:
            if command_line.options.wavelength is None:
                raise Usage("Can't get wavelength from image. Override with -w")
            wavelength = command_line.options.wavelength
        else:
            if command_line.options.wavelength is None:
                wavelength = beam.get_wavelength()
            else:
                wavelength = command_line.options.wavelength

        if beam is None and detector is None:
            if command_line.options.beam_center_x is None:
                print(
                    "Can't get beam x position from image. Using image center. Override with -x"
                )
                beam_x = raw_data.focus()[0] * pixel_size
            else:
                beam_x = command_line.options.beam_center_x * pixel_size

            if command_line.options.beam_center_y is None:
                print(
                    "Can't get beam y position from image. Using image center. Override with -y"
                )
                beam_y = raw_data.focus()[1] * pixel_size
            else:
                beam_y = command_line.options.beam_center_y * pixel_size
        else:
            if command_line.options.beam_center_x is None:
                beam_x = detector.get_beam_centre(beam.get_s0())[0]
            else:
                beam_x = command_line.options.beam_center_x * pixel_size

            if command_line.options.beam_center_y is None:
                beam_y = detector.get_beam_centre(beam.get_s0())[1]
            else:
                beam_y = command_line.options.beam_center_y * pixel_size

        if scan is None:
            timestamp = None
        else:
            msec, sec = math.modf(scan.get_epochs()[0])
            timestamp = serialtbx.util.timestamp((sec, msec))

        if is_multi_image:
            for i in range(img.get_num_images()):
                save_image(
                    command_line,
                    imgpath,
                    scan,
                    img.get_raw_data(i),
                    distance,
                    pixel_size,
                    wavelength,
                    beam_x,
                    beam_y,
                    overload,
                    timestamp,
                    image_number=i,
                )
        else:
            save_image(
                command_line,
                imgpath,
                scan,
                raw_data,
                distance,
                pixel_size,
                wavelength,
                beam_x,
                beam_y,
                overload,
                timestamp,
            )


def save_image(
    command_line,
    imgpath,
    scan,
    raw_data,
    distance,
    pixel_size,
    wavelength,
    beam_x,
    beam_y,
    overload,
    timestamp,
    image_number=None,
):
    if image_number is None:
        destpath = os.path.join(
            os.path.dirname(imgpath),
            os.path.splitext(os.path.basename(imgpath))[0] + ".pickle",
        )
    else:
        destpath = os.path.join(
            os.path.dirname(imgpath),
            os.path.splitext(os.path.basename(imgpath))[0]
            + "%05d.pickle" % image_number,
        )
    if command_line.options.skip_converted and os.path.isfile(destpath):
        if command_line.options.verbose:
            print("Skipping %s, file exists" % imgpath)
            return

    data = serialtbx.detector.cspad.dpack(
        data=raw_data,
        distance=distance,
        pixel_size=pixel_size,
        wavelength=wavelength,
        beam_center_x=beam_x,
        beam_center_y=beam_y,
        ccd_image_saturation=overload,
        saturated_value=overload,
        timestamp=timestamp,
    )

    if scan is not None:
        osc_start, osc_range = scan.get_oscillation()
        if osc_start != osc_range:
            data["OSC_START"] = osc_start
            data["OSC_RANGE"] = osc_range

            data["TIME"] = scan.get_exposure_times()[0]

    if command_line.options.crop:
        data = crop_image_pickle(data)

    if command_line.options.verbose:
        print("Writing", destpath)

    easy_pickle.dump(destpath, data)


if __name__ == "__main__":
    run()



from __future__ import annotations

import struct
import sys

from dxtbx.format.FormatTIFFRayonix import FormatTIFFRayonix


class FormatTIFFRayonixSPring8(FormatTIFFRayonix):
    """A class for reading TIFF format Rayonix images, and correctly
    constructing a model for the experiment from this, for SPring-8 beamlines
    with a reversed phi axis."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Rayonix TIFF format image,
        i.e. we can make sense of it. This simply checks that records which
        describe the size of the image match with the TIFF records which do
        the same."""

        width, height, depth, order, rawbytes = FormatTIFFRayonix.get_tiff_header(
            image_file
        )

        serial_number = -1

        for record in rawbytes[2464 : 2464 + 512].strip().split(b"\n"):
            if b"detector serial number" in record.lower():
                serial_number = int(record.split()[-1])

        # only understand a square image
        if width != height:
            return

        # BL26B2 MX225 with 1X1, 2X2, 3X3 or 4X4 binning
        if serial_number == 24 and width in [6144, 3072, 2046, 1536]:
            return True

        # BL32XU MX225-HE with 1X1, 2X2, 3X3 or 4X4 binning
        if serial_number == 31 and width in [6144, 3072, 2046, 1536]:
            return True

        # BL44XU MX225-HE with 1X1, 2X2, 3X3 or 4X4 binning
        if serial_number == 38 and width in [6144, 3072, 2046, 1536]:
            return True

        # BL44XU MX300-HE with 1X1, 2X2, 3X3 or 4X4 binning
        if serial_number == 42 and width in [8192, 4096, 2728, 2048]:
            return True

        # BL41XU MX225-HE with 1X1, 2X2, 3X3 or 4X4 binning
        if serial_number == 40 and width in [6144, 3072, 2046, 1536]:
            return True

        # BL32XU MX225-HS 1X1, 2X2, 3X3, 4X4, 5X5, 6X6, 8X8 or 10X10 binning
        if serial_number == 106 and width in [
            5760,
            2880,
            1920,
            1440,
            1152,
            960,
            720,
            576,
        ]:
            return True

        return False

    def _detector(self):
        """Return a model for a simple detector, which at the moment insists
        that the offsets and rotations are all 0.0."""

        starts, ends, offset, width = self._get_rayonix_scan_angles()
        rotations = self._get_rayonix_detector_rotations()

        # assert that two-theta offset is 0.0

        assert starts[0] == 0.0
        assert ends[0] == 0.0

        # assert that the rotations are all 0.0

        assert rotations[0] == 0.0
        assert rotations[1] == 0.0
        assert rotations[2] == 0.0

        distance = self._get_rayonix_distance()
        beam_x, beam_y = self._get_rayonix_beam_xy()
        pixel_size = self._get_rayonix_pixel_size()
        image_size = self._tiff_width, self._tiff_height
        max_trusted_value = struct.unpack(self._i, self._tiff_header_bytes[1128:1132])[
            0
        ]
        min_trusted_value = 0

        beam = beam_x * pixel_size[0], beam_y * pixel_size[1]

        for record in self._tiff_header_bytes[2464 : 2464 + 512].strip().split(b"\n"):
            if b"detector serial number" in record.lower():
                serial_number = int(record.split()[-1])
                break
        serial_to_model = {
            24: "225",
            31: "225HE",
            38: "225HE",
            42: "300HE",
            40: "225HE",
            106: "225HS",
        }
        model = serial_to_model[serial_number]
        gain = self._get_rayonix_gain(model)

        return self._detector_factory.simple(
            "CCD",
            distance,
            beam,
            "+x",
            "-y",
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
            [],
            gain=gain,
        )

    def _goniometer(self):
        """Return a model for goniometer corresponding to the values stored
        in the image header. In the first instance assume this is a single
        axis and raise exception otherwise."""

        self._get_rayonix_scan_angles()

        return self._goniometer_factory.single_axis_reverse()


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatTIFFRayonixSPring8.understand(arg))
        # print FormatTIFFRayonixSPring8BL26B2.understand(arg)



from __future__ import annotations

import os

from dxtbx.format.FormatPYunspecifiedStill import FormatPYunspecifiedStill
from dxtbx.model.experiment_list import ExperimentListFactory


def test_static_mask(dials_regression):
    filename = os.path.join(
        dials_regression,
        "image_examples/LCLS_CXI/shot-s00-2011-12-02T21_07Z29.723_00569.pickle",
    )
    assert FormatPYunspecifiedStill.understand(filename)

    expts_from_filename = ExperimentListFactory.from_filenames([filename])
    expts_from_dict = ExperimentListFactory.from_dict(expts_from_filename.to_dict())
    for expts in (expts_from_filename, expts_from_dict):
        assert len(expts) == 1
        imageset = expts[0].imageset
        assert imageset.get_format_class() == FormatPYunspecifiedStill
        mask = imageset.get_mask(0)
        assert len(mask) == 1
        assert mask[0].count(False) == 866781



from __future__ import annotations

import shutil
import subprocess

import pytest

import dxtbx


@pytest.mark.parametrize("use_mpi", [True, False])
def test_average(dials_data, tmp_path, use_mpi):
    # Only allow MPI tests if we've got MPI capabilities
    if use_mpi:
        pytest.importorskip("mpi4py")

    data = (
        dials_data("image_examples", pathlib=True) / "SACLA-MPCCD-run266702-0-subset.h5"
    )
    if use_mpi:
        command = "mpirun"
        mpargs = "-n 2 dxtbx.image_average --mpi=True".split()
    else:
        command = shutil.which("dxtbx.image_average")
        mpargs = "-n 2".split()
    result = subprocess.run(
        [command] + mpargs + "-v -a avg.cbf -s stddev.cbf -m max.cbf".split() + [data],
        cwd=tmp_path,
    )
    assert not result.returncode and not result.stderr

    h5 = dxtbx.load(data).get_detector()
    cbf = dxtbx.load(tmp_path / "avg.cbf").get_detector()

    assert h5.is_similar_to(cbf, ignore_trusted_range=True)
    assert h5[0].get_gain() == cbf[0].get_gain()



from __future__ import annotations

import os

import dxtbx


def test_noninteger_pedestal(dials_regression, tmpdir):
    filename = os.path.join(dials_regression, "image_examples/APS_14BMC/q315_1_001.img")
    # Read this file in as data
    with open(filename, "rb") as f:
        data = f.read()

    # Write out with an inserted header item of a noninteger pedestal
    test_file = tmpdir / "test_pedestal_001.img"
    with test_file.open("wb") as f:
        f.write(data.replace(b"DIM=2;\n", b"DIM=2;\nIMAGE_PEDESTAL=42.0;\n"))

    # Make sure this loads
    dxtbx.load(test_file)



from __future__ import annotations

import sys

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatus3AOS19ID6MSN132(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus3 images for 6M SN 132 @ APS19ID."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS3" in record
                and "S/N 60-0132" in header
            ):
                return True

        return False

    def _goniometer(self):
        """19ID has reversed goniometer"""

        return self._goniometer_factory.single_axis_reverse()


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatus3AOS19ID6MSN132.understand(arg))



from __future__ import annotations

import os

import pytest

from dxtbx.format.FormatNXmxDLS import FormatNXmxDLS
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.model.goniometer import Goniometer

pytest.importorskip("h5py")

master_h5 = (
    "/dls/mx/data/nt24686/nt24686-96/VMXi-AB5623/well_47/images/image_54771_master.h5"
)


@pytest.mark.skipif(
    not os.access(master_h5, os.R_OK), reason="Test images not available"
)
def test_VMXi_rotation_scan():
    assert FormatNXmxDLS.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatNXmxDLS

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == (0.075, 0.075)
    assert (
        tuple(reversed(panel.get_image_size())) == imageset[0][0].all() == (2162, 2068)
    )
    assert panel.get_trusted_range() == (0, 23315)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_origin() == pytest.approx((-79.155, 87.4875, -183.1849844439))
    assert panel.get_distance() == pytest.approx(183.1849844439)

    assert isinstance(gonio, Goniometer)
    assert gonio.get_rotation_axis() == (0, 1, 0)
    assert gonio.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert gonio.get_setting_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)

    assert scan.get_oscillation() == pytest.approx((-30, 0.1))
    assert scan.get_image_range() == (1, 600)

    assert beam.get_wavelength() == pytest.approx(0.7749042569928773)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))



from __future__ import annotations

import sys

try:
    import psana
except ImportError:
    psana = None
except TypeError:
    # Check if SIT_* environment variables are set
    import os

    if os.environ.get("SIT_ROOT"):
        # Variables are present, so must have been another error
        raise
    psana = None

from serialtbx.detector import rayonix

from libtbx.phil import parse
from scitbx.array_family import flex

from dxtbx.format.FormatXTC import FormatXTC, locator_str

rayonix_locator_str = """
  rayonix {
    bin_size = None
      .type = int
      .help = Detector binning mode
  }
"""

rayonix_locator_scope = parse(rayonix_locator_str + locator_str, process_includes=True)


class FormatXTCRayonix(FormatXTC):
    def __init__(self, image_file, **kwargs):
        super().__init__(image_file, locator_scope=rayonix_locator_scope, **kwargs)

        cfgs = self._ds.env().configStore()
        rayonix_cfg = cfgs.get(psana.Rayonix.ConfigV2, psana.Source("Rayonix"))
        assert rayonix_cfg.binning_f() == rayonix_cfg.binning_s()
        bin_size = rayonix_cfg.binning_f()
        if self.params.rayonix.bin_size is not None:
            assert bin_size == self.params.rayonix.bin_size
        self._pixel_size = rayonix.get_rayonix_pixel_size(bin_size)
        self._image_size = rayonix.get_rayonix_detector_dimensions(self._ds.env())

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(rayonix_locator_scope, image_file)
        except Exception:
            return False
        return any(["rayonix" in src.lower() for src in params.detector_address])

    def get_raw_data(self, index=None):
        if index is None:
            index = 0
        assert len(self.params.detector_address) == 1
        data = rayonix.get_data_from_psana_event(
            self._get_event(index), self.params.detector_address[0]
        )
        return flex.double(data)

    def get_detector(self, index=None):
        return self._detector()

    def _detector(self):
        return self._detector_factory.simple(
            sensor="UNKNOWN",
            distance=100.0,
            beam_centre=(50.0, 50.0),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self._pixel_size, self._pixel_size),
            image_size=self._image_size,
            trusted_range=(
                rayonix.rayonix_min_trusted_value,
                rayonix.rayonix_max_trusted_value,
            ),
            mask=[],
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXTCRayonix.understand(arg))



#!/usr/bin/env python
"""
Print out the contents of the dxtbx understanding of a bunch of images to
an example XDS.INP file. This should illustrate the usage of the dxtbx
classes.
"""


from __future__ import annotations

import math
import sys

from scitbx import matrix

from dxtbx.imageset import ImageSetFactory
from dxtbx.model.detector_helpers_types import detector_helpers_types


def xds_detector_name(xia2_name):
    """Translate from a xia2 name from the detector library to an XDS detector
    name."""

    if "pilatus" in xia2_name:
        return "PILATUS"
    if "rayonix" in xia2_name:
        return "CCDCHESS"
    if "adsc" in xia2_name:
        return "ADSC"
    if "saturn" in xia2_name:
        return "SATURN"
    if "raxis" in xia2_name:
        return "RAXIS"

    raise RuntimeError("detector %s unknown" % xia2_name)


class to_xds:
    def __init__(self, sequence):
        self._template = sequence.get_template()
        self._start_end = min(sequence.indices()), max(sequence.indices())
        self._goniometer = sequence.get_goniometer()
        self._detector = sequence.get_detector()
        self._beam = sequence.get_beam()
        self._scan = sequence.get_scan()

    def get_detector(self):
        return self._detector

    def get_goniometer(self):
        return self._goniometer

    def get_beam(self):
        return self._beam

    def get_scan(self):
        return self._scan

    def XDS(self):

        sensor = self.get_detector().get_type()
        fast, slow = map(int, self.get_detector().get_image_size())
        f, s = self.get_detector().get_pixel_size()
        df = int(1000 * f)
        ds = int(1000 * s)

        # FIXME probably need to rotate by pi about the X axis

        R = matrix.col((1.0, 0.0, 0.0)).axis_and_angle_as_r3_rotation_matrix(
            180.0, deg=True
        )

        detector = xds_detector_name(
            detector_helpers_types.get(sensor, fast, slow, df, ds)
        )
        trusted = self.get_detector().get_trusted_range()

        # XDS OVERLOAD parameter is the maximum trusted value, not the first
        # overloaded value (https://xds.mr.mpg.de/html_doc/xds_parameters.html#OVERLOAD=)
        print(
            "DETECTOR=%s MINIMUM_VALID_PIXEL_VALUE=%d OVERLOAD=%d"
            % (detector, trusted[0], trusted[1])
        )

        if detector == "PILATUS":
            print("SENSOR_THICKNESS= 0.32")

        print(
            "DIRECTION_OF_DETECTOR_X-AXIS= %.3f %.3f %.3f"
            % (R * self.get_detector().get_fast_axis()).elems
        )

        print(
            "DIRECTION_OF_DETECTOR_Y-AXIS= %.3f %.3f %.3f"
            % (R * self.get_detector().get_slow_axis()).elems
        )

        print("NX=%d NY=%d QX=%.4f QY=%.4f" % (fast, slow, f, s))

        F = R * self.get_detector().get_fast_axis()
        S = R * self.get_detector().get_slow_axis()
        N = F.cross(S)

        origin = R * self.get_detector().get_origin()
        beam = (
            R
            * self.get_beam().get_sample_to_source_direction()
            / math.sqrt(
                matrix.col(self.get_beam().get_sample_to_source_direction()).dot()
            )
        )
        centre = -(origin - origin.dot(N) * N)
        x = centre.dot(F)
        y = centre.dot(S)

        print("DETECTOR_DISTANCE= %.3f" % origin.dot(N))
        print("ORGX= %.1f ORGY= %.1f" % (x / f, y / s))
        print(
            "ROTATION_AXIS= %.3f %.3f %.3f"
            % (R * self.get_goniometer().get_rotation_axis()).elems
        )
        print("STARTING_ANGLE= %.3f" % self.get_scan().get_oscillation()[0])
        print("OSCILLATION_RANGE= %.3f" % self.get_scan().get_oscillation()[1])
        print("X-RAY_WAVELENGTH= %.5f" % self.get_beam().get_wavelength())
        print("INCIDENT_BEAM_DIRECTION= %.3f %.3f %.3f" % (-beam).elems)
        print(
            "FRACTION_OF_POLARIZATION= %.3f"
            % self.get_beam().get_polarization_fraction()
        )
        print(
            "POLARIZATION_PLANE_NORMAL= %.3f %.3f %.3f"
            % self.get_beam().get_polarization_normal()
        )
        print("NAME_TEMPLATE_OF_DATA_FRAMES= %s" % self._template.replace("#", "?"))
        print("TRUSTED_REGION= 0.0 1.41")
        for f0, f1, s0, s1 in self.get_detector().get_mask():
            print("UNTRUSTED_RECTANGLE= %d %d %d %d" % (f0, f1 + 1, s0, s1 + 1))

        start_end = self.get_scan().get_image_range()

        if start_end[0] == 0:
            start_end = (1, start_end[1])

        print("DATA_RANGE= %d %d" % start_end)
        print("JOB=XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT")


def factory(list_of_images):
    sequences = ImageSetFactory.new(list_of_images)
    assert len(sequences) == 1
    return sequences[0]


if __name__ == "__main__":
    # run some tests

    xsx = to_xds(factory(sys.argv[1:]))
    xsx.XDS()



"""A simple cache controller. Caching only one file at a time."""


from __future__ import annotations

import os
import threading

import dxtbx.filecache


class simple_controller:
    """A simple cache controller. Caching one file at a time."""

    def __init__(self):
        # Reference to the identifier of the last seen object
        # and the relevant cache object
        self._cache_tag = None
        self._cache = None

        # Lock for concurrent access
        self._lock = threading.Lock()

        # Keep the current PID to detect the use of multiprocessing parallelization
        # which breaks caching assumptions
        self._pid = os.getpid()

    def __del__(self):
        """Garbage collection.
        Since only a cache controller can run .open() on lazy cache objects,
        the lazy cache object can be told to close as soon as possible.
        File handles can still be open legitimately."""
        try:
            if self._cache is not None and self._pid == os.getpid():
                self._cache.close()
        except (AttributeError, TypeError):
            pass
        self._cache = None

    def check(self, tag, open_method):
        """The main cache controller access method. Checks if an object with name
        "tag" is cached. If so, returns a (pseudo-, ie. cached) file handle to
        this object.
        Otherwise, create a cache first, using the passed open_method()
        function, which returns a (true) file handle."""
        with self._lock:
            currentpid = os.getpid()
            if currentpid != self._pid:
                #       print "(%5s) Controller %s detected PID change, flushing %s" % (os.getpid(), format(id(self), '#x'), self._cache_tag)
                # Change the PID and drop reference to the previous cache.
                self._cache = None
                self._cache_tag = None
                self._pid = currentpid
                # NB: Explicitly do not close the cache in this case. Would be pointless.

            # If tag has changed then we are changing cache contents
            if tag != self._cache_tag:
                if self._cache is not None:
                    #         print "(%5s) Controller %s closing previous cache on %s" % (os.getpid(), format(id(self), '#x'), self._cache_tag)
                    try:
                        self._cache.close()
                    finally:
                        self._cache = None
                        self._cache_tag = None

                #       print "(%5s) Controller %s opening cache on %s" % (os.getpid(), format(id(self), '#x'), self._tag)
                try:
                    self._cache_tag = tag
                    self._cache = dxtbx.filecache.lazy_file_cache(open_method())
                except:  # noqa: E722
                    # Make sure we leave in a valid state
                    self._cache_tag = None
                    self._cache = None
                    raise
            #     else:
            #       print "(%5s) Controller %s reporting cache hit on %s" % (os.getpid(), format(id(self), '#x'), self._cache_tag)
            return self._cache.open()


class non_caching_controller:
    """A controller that does not do any caching."""

    @staticmethod
    def check(tag, open_method):
        """Open the file. No caching required."""
        print("Caching disabled for", tag)
        return open_method()


# To disable all caching uncomment the following line:
# simple_controller = non_caching_controller



from __future__ import annotations

import math

from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.model import MultiAxisGoniometer

try:
    from ..dxtbx_masking_ext import (
        GoniometerShadowMasker,
        SmarGonShadowMasker,
        is_inside_polygon,
        mask_untrusted_circle,
        mask_untrusted_polygon,
        mask_untrusted_rectangle,
    )
except ModuleNotFoundError:
    from dxtbx_masking_ext import (  # type: ignore
        GoniometerShadowMasker,
        SmarGonShadowMasker,
        is_inside_polygon,
        mask_untrusted_circle,
        mask_untrusted_polygon,
        mask_untrusted_rectangle,
    )

__all__ = [
    "GoniometerShadowMasker",
    "SmarGonShadowMasker",
    "is_inside_polygon",
    "mask_untrusted_circle",
    "mask_untrusted_polygon",
    "mask_untrusted_rectangle",
]


class GoniometerMaskerFactory:
    @staticmethod
    def mini_kappa(goniometer, cone_opening_angle=43.60281897270362):
        """Construct a GoniometerShadowMasker for a mini-kappa goniometer.

        This is modelled a simple cone with the opening angle specified by
        `cone_opening_angle`.

        Args:
            goniometer (`dxtbx.model.Goniometer`): The goniometer instance.
            cone_opening_angle (float): The opening angle of the cone (in degrees).

        Returns:
            `dxtbx.masking.GoniometerShadowMasker`

        """
        assert isinstance(goniometer, MultiAxisGoniometer)
        assert len(goniometer.get_axes()) == 3
        # Simple model of cone around goniometer phi axis
        # Exact values don't matter, only the ratio of height/radius
        height = 50  # mm

        radius_height_ratio = math.tan(1 / 2 * cone_opening_angle * math.pi / 180)
        radius = radius_height_ratio * height

        steps_per_degree = 1
        theta = (
            flex.double(range(360 * steps_per_degree))
            * math.pi
            / 180
            * 1
            / steps_per_degree
        )
        y = radius * flex.cos(-theta)  # x
        z = radius * flex.sin(-theta)  # y
        x = flex.double(theta.size(), height)  # z

        coords = flex.vec3_double(zip(x, y, z))
        coords.insert(0, (0, 0, 0))

        return GoniometerShadowMasker(goniometer, coords, flex.size_t(len(coords), 0))

    @staticmethod
    def dls_i23_kappa(goniometer):
        """Construct a GoniometerShadowMasker for the DLS I23 Kappa goniometer.

        Args:
            goniometer (`dxtbx.model.Goniometer`): The goniometer instance.

        Returns:
            `dxtbx.masking.GoniometerShadowMasker`

        """
        coords = flex.vec3_double(((0, 0, 0),))

        alpha = flex.double_range(0, 190, step=10) * math.pi / 180
        r = flex.double(alpha.size(), 40)
        x = flex.double(r.size(), 107.61)
        y = -r * flex.sin(alpha)
        z = -r * flex.cos(alpha)
        coords.extend(flex.vec3_double(x, y, z))

        coords.extend(
            flex.vec3_double(
                (
                    # fixed
                    (107.49, 7.84, 39.49),
                    (107.39, 15.69, 38.97),
                    (107.27, 23.53, 38.46),
                    (107.16, 31.37, 37.94),
                    (101.76, 33.99, 36.25),
                    (96.37, 36.63, 34.56),
                    (90.98, 39.25, 33.00),
                    (85.58, 41.88, 31.18),
                    (80.89, 47.06, 31.00),
                    (76.55, 51.51, 31.03),
                    (72.90, 55.04, 31.18),
                    (66.86, 60.46, 31.67),
                    (62.10, 64.41, 32.25),
                )
            )
        )

        alpha = flex.double_range(180, 370, step=10) * math.pi / 180
        r = flex.double(alpha.size(), 33)
        x = flex.sqrt(flex.pow2(r * flex.sin(alpha)) + 89.02**2) * flex.cos(
            (50 * math.pi / 180) - flex.atan(r / 89.02 * flex.sin(alpha))
        )
        y = flex.sqrt(flex.pow2(r * flex.sin(alpha)) + 89.02**2) * flex.sin(
            (50 * math.pi / 180) - flex.atan(r / 89.02 * flex.sin(alpha))
        )
        z = -r * flex.cos(alpha)
        coords.extend(flex.vec3_double(x, y, z))

        coords.extend(
            flex.vec3_double(
                (
                    # fixed
                    (62.10, 64.41, -32.25),
                    (66.86, 60.46, -31.67),
                    (72.90, 55.04, -31.18),
                    (76.55, 51.51, -31.03),
                    (80.89, 47.06, -31.00),
                    (85.58, 41.88, -31.18),
                    (90.98, 39.25, -33.00),
                    (96.37, 36.63, -34.56),
                    (101.76, 33.99, -36.25),
                    (107.16, 31.37, -37.94),
                    (107.27, 23.53, -38.46),
                    (107.39, 15.69, -38.97),
                    (107.49, 7.84, -39.49),
                    (107.61, 0.00, -40.00),
                )
            )
        )

        # I23 end station coordinate system:
        #   X-axis: positive direction is facing away from the storage ring (from
        #           sample towards goniometer)
        #   Y-axis: positive direction is vertically up
        #   Z-axis: positive direction is in the direction of the beam (from
        #           sample towards detector)
        #   K-axis (kappa): at an angle of +50 degrees from the X-axis
        #   K & phi rotation axes: clockwise rotation is positive (right hand
        #           thumb rule)
        #   Omega-axis: along the X-axis; clockwise rotation is positive

        # End station x-axis is parallel to ImgCIF x-axis
        # End station z-axis points in opposite direction to ImgCIF definition
        # (ImgCIF: The Z-axis is derived from the source axis which goes from
        # the sample to the source)
        # Consequently end station y-axis (to complete set following right hand
        # rule) points in opposite direction to ImgCIF y-axis.
        # Kappa arm aligned with -y in ImgCIF convention

        R = align_reference_frame(
            matrix.col((1, 0, 0)),
            matrix.col((1, 0, 0)),
            matrix.col((0, 1, 0)),
            matrix.col((0, -1, 0)),
        )
        coords = R.elems * coords

        return GoniometerShadowMasker(goniometer, coords, flex.size_t(len(coords), 1))

    @staticmethod
    def smargon(goniometer):
        """Construct a SmarGonShadowMasker for the SmarGon goniometer.

        Args:
            goniometer (`dxtbx.model.Goniometer`): The goniometer instance.

        Returns:
            `dxtbx.masking.SmarGonShadowMasker`

        """
        return SmarGonShadowMasker(goniometer)

    @staticmethod
    def diamond_anvil_cell(goniometer, cone_opening_angle):
        radius_height_ratio = math.tan(1 / 2 * cone_opening_angle)
        height = 10  # mm
        radius = radius_height_ratio * height

        steps_per_degree = 1
        theta = (
            flex.double([list(range(360 * steps_per_degree))])
            * math.pi
            / 180
            * 1
            / steps_per_degree
        )
        x = radius * flex.cos(theta)  # x
        z = radius * flex.sin(theta)  # y
        y = flex.double(theta.size(), height)  # z

        coords = flex.vec3_double(zip(x, y, z))
        coords.extend(flex.vec3_double(zip(x, -y, z)))
        coords.insert(0, (0, 0, 0))

        return GoniometerShadowMasker(
            goniometer, coords, flex.size_t(len(coords), 0), True
        )



from __future__ import annotations

import sys

from iotbx.detectors.bruker import BrukerImage
from scitbx import matrix

from dxtbx.format.FormatBruker import FormatBruker


class FormatBrukerFixedChi(FormatBruker):
    @staticmethod
    def understand(image_file):

        hdr = FormatBruker.read_header_lines(image_file)
        hdr_dic = FormatBruker.parse_header(hdr)

        if "FixedChiStage" not in hdr_dic["MODEL"]:
            return False
        if "PHOTONII" in hdr_dic["DETTYPE"]:
            return False

        return True

    def _start(self):
        self.header_dict = {}
        image_blob = open(self._image_file, "rb").read()
        header_text = image_blob.split(b"......")[0].decode()
        for j in range(0, len(header_text), 80):
            record = header_text[j : j + 80]
            if record.startswith("CFR:"):
                continue
            key = record[:8]
            assert key[-1] == ":"
            values = record[8:]
            self.header_dict[key.replace(":", "").strip()] = [
                v.strip() for v in values.split()
            ]

        self.detectorbase = BrukerImage(self._image_file)

    def _goniometer(self):
        # goniometer angles in ANGLES are 2-theta, omega, phi, chi (FIXED)
        # AXIS indexes into this list to define the scan axis (in FORTRAN counting)
        # START and RANGE define the start and step size for each image
        # assume omega is 1,0,0 axis; chi about 0,0,1 at datum
        angles = [float(a) for a in self.header_dict["ANGLES"]]

        beam = matrix.col((0, 0, 1))
        phi = matrix.col((1, 0, 0)).rotate_around_origin(-beam, angles[3], deg=True)

        if self.header_dict["AXIS"][0] == "2":
            # OMEGA scan
            axis = (-1, 0, 0)
            incr = float(self.header_dict["INCREME"][0])
            if incr < 0:
                axis = (1, 0, 0)
            fixed = phi.axis_and_angle_as_r3_rotation_matrix(angles[2], deg=True)
            return self._goniometer_factory.make_goniometer(axis, fixed.elems)
        else:
            # PHI scan
            assert self.header_dict["AXIS"][0] == "3"
            omega = matrix.col((1, 0, 0))
            axis = phi.rotate_around_origin(omega, angles[1], deg=True)
            return self._goniometer_factory.known_axis(axis.elems)

    def _detector(self):
        # goniometer angles in ANGLES are 2-theta, omega, phi, chi (FIXED)
        two_theta = float(self.header_dict["ANGLES"][0])
        max_trusted_value = 60000
        min_trusted_value = 0

        fast = matrix.col((1, 0, 0))
        slow = matrix.col((0, 1, 0))
        beam = matrix.col((0, 0, 1))
        pixel_mm = 5.0 / float(self.header_dict["DETTYPE"][1])
        beam_pixel = [float(bp) for bp in self.header_dict["CENTER"][:2]]
        distance_mm = 10.0 * float(self.header_dict["DISTANC"][1])
        origin = (
            -distance_mm * beam
            - fast * pixel_mm * beam_pixel[1]
            - slow * pixel_mm * beam_pixel[0]
        )
        origin = origin.rotate_around_origin(-fast, two_theta, deg=True)
        slow = slow.rotate_around_origin(-fast, two_theta, deg=True)
        pixel_size = pixel_mm, pixel_mm
        image_size = (
            int(self.header_dict["NROWS"][0]),
            int(self.header_dict["NCOLS"][0]),
        )

        return self._detector_factory.complex(
            "CCD",
            origin.elems,
            fast.elems,
            slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
        )

    def _beam(self):
        wavelength = float(self.header_dict["WAVELEN"][0])

        return self._beam_factory.simple(wavelength)

    def _scan(self):

        start = float(self.header_dict["START"][0])
        incr = float(self.header_dict["INCREME"][0])
        if incr < 0:
            start *= -1
            incr *= -1

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=1,
            osc_start=start,
            osc_width=incr,
            epoch=None,
        )

    # FIXME implement get_raw_data including LINEAR factor (multiplier) =>
    # explains the observed GAIN around 10


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatBrukerFixedChi.understand(arg))



# LIBTBX_SET_DISPATCHER_NAME dev.dxtbx.read_sequence

"""Tool to benchmark overall time cost for simply reading data"""

from __future__ import annotations

import argparse
import time

import dxtbx.util
from dxtbx.imageset import ImageSetFactory


def read_sequence(images: list[str]):

    sequences = ImageSetFactory.new(images)

    for sequence in sequences:
        print(sequence.get_detector())
        print(sequence.get_scan())

        indices = sequence.indices()

        t0 = time.time()
        for i in indices:
            sequence.get_raw_data(i)
        t1 = time.time()

        print(f"Reading {len(indices)} frames took {t1-t0:.2f}s")


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(
        description="Benchmark the time to read a set of images"
    )
    parser.add_argument("images", metavar="IMAGE", help="Images to read", nargs="+")
    options = parser.parse_args(args)
    return read_sequence(options.images)


if __name__ == "__main__":
    run()



from __future__ import annotations

import sys

import numpy as np

from cctbx import factor_kev_angstrom
from cctbx.eltbx import attenuation_coefficient
from libtbx.phil import parse
from scitbx.array_family import flex
from scitbx.matrix import col

from dxtbx.format.FormatXTC import FormatXTC, locator_str
from dxtbx.model import Detector, ParallaxCorrectedPxMmStrategy

epix_locator_str = """
"""

epix_locator_scope = parse(epix_locator_str + locator_str, process_includes=True)


class FormatXTCEpix(FormatXTC):
    def __init__(self, image_file, **kwargs):
        super().__init__(image_file, locator_scope=epix_locator_scope, **kwargs)
        self._cached_detector = {}

    @staticmethod
    def understand(image_file):
        try:
            params = FormatXTC.params_from_phil(epix_locator_scope, image_file)
        except Exception:
            return False
        return any(["epix" in src.lower() for src in params.detector_address])

    def get_raw_data(self, index=None):
        if index is None:
            index = 0
        d = FormatXTCEpix.get_detector(self, index)
        evt = self._get_event(index)
        run = self.get_run_from_index(index)
        det = self._get_psana_detector(run)
        data = det.calib(evt)
        data = data.astype(np.float64)
        # the shape of the epix 10k data is (16, 352, 384)

        self._raw_data = []
        for module_count, module in enumerate(d.hierarchy()):
            for asic_count, asic in enumerate(module):
                fdim, sdim = asic.get_image_size()
                sensor_id = asic_count // 2  # There are 2X2 asics per module
                asic_in_sensor_id = asic_count % 2  # this number will be 0 or 1
                asic_data = data[module_count][
                    sensor_id * sdim : (sensor_id + 1) * sdim,
                    asic_in_sensor_id * fdim : (asic_in_sensor_id + 1) * fdim,
                ]  # 8 sensors per module
                self._raw_data.append(flex.double(np.array(asic_data)))
        assert len(d) == len(self._raw_data), (len(d), len(self._raw_data))
        return tuple(self._raw_data)

    def get_detector(self, index=None):
        return FormatXTCEpix._detector(self, index)

    def _detector(self, index=None):
        from PSCalib.SegGeometryStore import sgs
        from serialtbx.detector.xtc import basis_from_geo

        run = self.get_run_from_index(index)
        if run.run() in self._cached_detector:
            return self._cached_detector[run.run()]

        if index is None:
            index = 0
        assert len(self.params.detector_address) == 1

        wavelength = self.get_beam(index).get_wavelength()

        det = self._get_psana_detector(run)

        geom = det.pyda.geoaccess(self._get_event(index).run())
        pixel_size = det.pixel_size(self._get_event(index)) / 1000.0  # convert to mm
        d = Detector()
        pg0 = d.hierarchy()
        # first deal with D0
        det_num = 0
        root = geom.get_top_geo()
        root_basis = basis_from_geo(root)
        while len(root.get_list_of_children()) == 1:
            sub = root.get_list_of_children()[0]
            sub_basis = basis_from_geo(sub)
            root = sub
            root_basis = root_basis * sub_basis
        t = root_basis.translation
        root_basis.translation = col((t[0], t[1], -abs(t[2])))

        origin = col((root_basis * col((0, 0, 0, 1)))[0:3])
        fast = col((root_basis * col((1, 0, 0, 1)))[0:3]) - origin
        slow = col((root_basis * col((0, 1, 0, 1)))[0:3]) - origin
        pg0.set_local_frame(fast.elems, slow.elems, origin.elems)
        pg0.set_name("D%d" % (det_num))

        # Modules next
        for module_num in range(len(root.get_list_of_children())):
            pg1 = pg0.add_group()
            module = root.get_list_of_children()[module_num]
            module_basis = basis_from_geo(module)
            origin = col((module_basis * col((0, 0, 0, 1)))[0:3])
            fast = col((module_basis * col((1, 0, 0, 1)))[0:3]) - origin
            slow = col((module_basis * col((0, 1, 0, 1)))[0:3]) - origin
            pg1.set_local_frame(fast.elems, slow.elems, origin.elems)
            pg1.set_name("D%dM%d" % (det_num, module_num))

            # Read the known layout of the Epix 2x2 module
            sg = sgs.Create(segname=module.oname)
            xx, yy = sg.get_seg_xy_maps_um()
            xx = xx / 1000
            yy = yy / 1000

            # Now deal with ASICs
            for asic_num in range(4):
                val = "ARRAY_D0M%dA%d" % (module_num, asic_num)
                p = pg1.add_panel()
                dim_slow = xx.shape[0]
                dim_fast = xx.shape[1]
                sensor_id = asic_num // 2  # There are 2X2 asics per module
                asic_in_sensor_id = asic_num % 2  # this number will be 0 or 1
                id_slow = sensor_id * (dim_slow // 2)
                id_fast = asic_in_sensor_id * (dim_fast // 2)
                origin = col((xx[id_slow][id_fast], yy[id_slow][id_fast], 0))
                fp = col((xx[id_slow][id_fast + 1], yy[id_slow][id_fast + 1], 0))
                sp = col((xx[id_slow + 1][id_fast], yy[id_slow + 1][id_fast], 0))
                fast = (fp - origin).normalize()
                slow = (sp - origin).normalize()
                p.set_local_frame(fast.elems, slow.elems, origin.elems)
                p.set_pixel_size((pixel_size, pixel_size))
                p.set_image_size((dim_fast // 2, dim_slow // 2))
                p.set_trusted_range((0, 2e6))
                p.set_gain(factor_kev_angstrom / wavelength)
                p.set_name(val)

        try:
            beam = self._beam(index)
        except Exception:
            print(
                "No beam object initialized. Returning ePix detector without parallax corrections"
            )
            self._cached_detector[run.run()] = d
            return d

        # take into consideration here the thickness of the sensor also the
        # wavelength of the radiation (which we have in the same file...)
        wavelength = beam.get_wavelength()
        thickness = 0.5  # mm

        material = "Si"
        table = attenuation_coefficient.get_table(material)
        # mu_at_angstrom returns cm^-1
        mu = table.mu_at_angstrom(wavelength) / 10.0  # mu: mm^-1
        t0 = thickness
        for panel in d:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
            panel.set_thickness(thickness)
            panel.set_material(material)
            panel.set_mu(mu)

        self._cached_detector[run.run()] = d
        return d


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        # Bug, should call this part differently for understand method to work
        print(FormatXTCEpix.understand(arg))



"""
Base implementation of fullCBF format - as used with Dectris detectors
amongst others - this will read the header and construct the full model,
but will allow for extension for specific implementations of CBF.
"""


from __future__ import annotations

import sys

import numpy as np
import pycbf

from iotbx.detectors.cbf import CBFImage
from scitbx.array_family import flex

from dxtbx.format.Format import bz2, gzip
from dxtbx.format.FormatCBF import FormatCBF
from dxtbx.format.FormatStill import FormatStill

# Does the pybf library expose read_buffer?
if hasattr(pycbf.cbf_handle_struct, "read_buffer"):

    def cbf_read_buffer(handle, buffer, flags):
        return handle.read_buffer(buffer, flags)

else:
    from dxtbx.format.image import cbf_read_buffer


class FormatCBFFull(FormatCBF):
    """An image reading class for full CBF format images i.e. those from
    a variety of cameras which support this format."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBF.get_cbf_header(image_file)

        if "_diffrn.id" not in header and "_diffrn_source" not in header:
            return False

        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        # It appears Pycbf can not handle unicode filenames (see dials/dials#256)
        self._raw_data = None
        super().__init__(str(image_file), **kwargs)

    def __del__(self):
        self._cbf_handle.__swig_destroy__(self._cbf_handle)

    def _get_cbf_handle(self):
        try:
            return self._cbf_handle
        except AttributeError:
            self._cbf_handle = pycbf.cbf_handle_struct()

            buffer = None
            # Reopen to tell if it's a gzip - this should be cached, so fast
            with FormatCBF.open_file(self._image_file, "rb") as fin:
                # If this was a gzip or bzip file, read as a buffer
                if (gzip and isinstance(fin._cache_object._file, gzip.GzipFile)) or (
                    bz2 and isinstance(fin._cache_object._file, bz2.BZ2File)
                ):
                    buffer = fin.read()

            if buffer:
                cbf_read_buffer(self._cbf_handle, buffer, pycbf.MSG_DIGEST)
            else:
                self._cbf_handle.read_widefile(
                    self._image_file.encode(), pycbf.MSG_DIGEST
                )

            return self._cbf_handle

    def _goniometer(self):
        """Return a working goniometer instance."""

        return self._goniometer_factory.imgCIF_H(self._get_cbf_handle())

    def _detector(self):
        """Return a working detector instance."""

        return self._detector_factory.imgCIF_H(self._get_cbf_handle(), "unknown")

    def _beam(self):
        """Return a working beam instance."""

        return self._beam_factory.imgCIF_H(self._get_cbf_handle())

    def _scan(self):
        """Return a working scan instance."""

        return self._scan_factory.imgCIF_H(self._image_file, self._get_cbf_handle())

    def detectorbase_start(self):
        self.detectorbase = CBFImage(self._image_file)
        self.detectorbase.readHeader()

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array."""
        if self._raw_data is not None:
            return self._raw_data

        self.detectorbase_start()
        try:
            image = self.detectorbase
            image.read()
            self._raw_data = image.get_raw_data()

            return self._raw_data
        except Exception:
            return None


class FormatCBFFullStill(FormatStill, FormatCBFFull):
    """An image reading class for full CBF format images i.e. those from
    a variety of cameras which support this format. Custom derived from
    the FormatStill to handle images without a gonimeter or scan"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBF.get_cbf_header(image_file)

        if "_diffrn.id" not in header and "_diffrn_source" not in header:
            return False

        # According to ImageCIF, "Data items in the DIFFRN_MEASUREMENT_AXIS
        # category associate axes with goniometers."
        # http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Cdiffrn_measurement_axis.html
        if "diffrn_measurement_axis" in header:
            return False

        # This implementation only supports single panel.
        try:
            cbf_handle = pycbf.cbf_handle_struct()
            cbf_handle.read_widefile(image_file.encode(), pycbf.MSG_DIGEST)
        except Exception as e:
            if "CBFlib Error" in str(e):
                return False

        # check if multiple arrays
        try:
            return cbf_handle.count_elements() == 1
        except Exception as e:
            if "CBFlib Error" in str(e):
                return False

        return True

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array."""
        if self._raw_data is not None:
            return self._raw_data

        # Override parent's get_raw_data, which relies on iotbx, which in turn
        # relies on cbflib_adaptbx, which in turn expects a gonio
        cbf = self._get_cbf_handle()

        cbf.find_category(b"array_structure")
        cbf.find_column(b"encoding_type")
        cbf.select_row(0)
        types = []
        for i in range(cbf.count_rows()):
            types.append(cbf.get_value())
            cbf.next_row()
        assert (
            len(types) == cbf.count_rows() == 1
        )  # multi-tile data read by different class
        dtype = types[0]

        # find the data
        cbf.select_category(0)
        while cbf.category_name().lower() != b"array_data":
            try:
                cbf.next_category()
            except Exception:
                return None
        cbf.select_column(0)
        cbf.select_row(0)

        cbf.find_column(b"data")
        assert cbf.get_typeofvalue().find(b"bnry") > -1

        # handle floats vs ints
        if dtype == b"signed 32-bit integer":
            array_string = cbf.get_integerarray_as_string()
            self._raw_data = flex.int(np.fromstring(array_string, np.int32))
            parameters = cbf.get_integerarrayparameters_wdims_fs()
            slow, mid, fast = (parameters[11], parameters[10], parameters[9])
            assert slow == 1  # sections not supported
            array_size = mid, fast
        elif dtype == b"signed 64-bit real IEEE":
            array_string = cbf.get_realarray_as_string()
            self._raw_data = flex.double(np.fromstring(array_string, float))
            parameters = cbf.get_realarrayparameters_wdims_fs()
            slow, mid, fast = (parameters[7], parameters[6], parameters[5])
            assert slow == 1  # sections not supported
            array_size = mid, fast
        else:
            return None  # type not supported

        self._raw_data.reshape(flex.grid(*array_size))
        return self._raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFFull.understand(arg))



from __future__ import annotations

import io
import os
import random
from pathlib import Path

import dxtbx.filecache


def test_filecache():
    image = Path(__file__).parent / "phi_scan_001.cbf"
    correct_data = image.read_bytes()

    # Create a caching object
    cache = dxtbx.filecache.lazy_file_cache(open(image, "rb"))

    # read 100 bytes
    sh = io.BytesIO(correct_data)
    with cache.open() as fh:
        actual = fh.read(100)
        expected = sh.read(100)
        assert actual == expected
        actual = fh.read(0)
        expected = sh.read(0)
        assert actual == expected
        actual = fh.read(5000)
        expected = sh.read(5000)
        assert actual == expected

    # readlines
    sh = io.BytesIO(correct_data)
    with cache.open() as fh:
        actual = fh.readlines()
        expected = sh.readlines()
        assert actual == expected

    # 5x readline
    sh = io.BytesIO(correct_data)
    with cache.open() as fh:
        actual = [fh.readline() for n in range(5)]
        expected = [sh.readline() for n in range(5)]
        assert actual == expected

    # Get a new cache object
    cache.close()
    cache = dxtbx.filecache.lazy_file_cache(open(image, "rb"))
    cache._page_size = 5
    fh = dxtbx.filecache.pseudo_file(cache)

    # readline stress test
    sh = io.BytesIO(correct_data)
    with cache.open() as fh:
        actual = fh.readline()
        expected = sh.readline()
        assert actual == expected

        actual = fh.read(68)
        expected = sh.read(68)
        assert actual == expected

        actual = fh.readline()
        expected = sh.readline()
        assert actual == expected

        actual = fh.read(1)
        expected = sh.read(1)
        assert actual == expected

    # Get a new cache object
    cache.close()
    cache = dxtbx.filecache.lazy_file_cache(open(image, "rb"))

    sh = io.BytesIO(correct_data)
    fh = dxtbx.filecache.pseudo_file(cache)

    random_a, random_b = random.randint(0, 10000), random.randint(0, 150000)
    print("Running test for parameters %d %d" % (random_a, random_b))

    actual = (fh.readline(), fh.read(random_a), fh.readline(), fh.read(random_b))
    expected = (sh.readline(), sh.read(random_a), sh.readline(), sh.read(random_b))
    assert actual == expected


def test_filecache_more(dials_regression):
    filename = os.path.join(
        dials_regression, "image_examples", "MacScience", "reallysurprise_001.ipf"
    )
    cache = dxtbx.filecache.lazy_file_cache(open(filename, "rb"))
    with dxtbx.filecache.pseudo_file(cache) as fh:
        fh.seek(3000 * 3000 * 2)
        fh.read(1024)
        assert fh.read(1) == b""

    cache = dxtbx.filecache.lazy_file_cache(open(filename, "rb"))
    with dxtbx.filecache.pseudo_file(cache) as fh:
        data = fh.read(3000 * 3000 * 2)
        assert len(data) == 3000 * 3000 * 2
        assert len(fh.read(1024)) == 1024
        assert fh.read(1) == b""

    cache = dxtbx.filecache.lazy_file_cache(open(filename, "rb"))
    with dxtbx.filecache.pseudo_file(cache) as fh:
        fh.read(1024)
        data = fh.read()
        assert len(data) == 3000 * 3000 * 2


def test_read_termination_at_end_of_file(tmpdir):
    testfile = tmpdir.join("test")
    testfile.write("\n".join(str(n) for n in range(5)))
    cache = dxtbx.filecache.lazy_file_cache(testfile.open("rb"))
    with cache.open() as fh:
        for record in fh:
            assert record, "Loop should have terminated already"



"""
Sub class of FormatTIFFRayonix specialized for the XPP Rayonix detector at LCLS

Images from the XPP Rayonix detector have several uninitialized values, such as
distance, wavelength, etc.  Set these values to zero so the images can be at
least viewed.
"""


from __future__ import annotations

import re
import struct
import sys

from iotbx.detectors.mar import MARImage

from dxtbx.format.FormatTIFFRayonix import FormatTIFFRayonix


def check(l):
    """Sets l or values in l that are less than zero to zero"""
    if not isinstance(l, list) and not isinstance(l, tuple):
        if l < 0:
            return 0
        else:
            return l
    ret = []
    for val in l:
        if isinstance(val, list):
            ret.append(check(val))
        else:
            if val < 0:
                val = 0
            ret.append(val)

    if isinstance(l, tuple):
        return tuple(ret)
    return ret


_xpp_pattern = re.compile(r".*xpp[a-zA-Z][0-9][0-9][0-9][0-9].*")


class FormatTIFFRayonixXPP(FormatTIFFRayonix):
    """A class for reading TIFF format Rayonix images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an XPP Rayonix TIFF"""

        with open(image_file, "rb") as fh:

            def get(index, typelen, offset=1024):
                fh.seek(offset + index)
                rawdata = fh.read(typelen)
                return struct.unpack("<c", rawdata)[0]

            data = [get(1152 + i, 1) for i in range(128)]
        filepath = b"".join(c for c in data if 32 < ord(c) < 127).decode(
            "latin-1", "replace"
        )

        return bool(_xpp_pattern.match(filepath))

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        MARImage._read_header_asserts = lambda self: None
        super().__init__(image_file, **kwargs)

    ####################################################################
    #                                                                  #
    # Helper methods to get all of the values out of the TIFF header   #
    # - separated out to assist with code clarity                      #
    #                                                                  #
    ####################################################################

    def _get_rayonix_beam_xy(self):
        """Get the beam x, y positions which are defined in the standard
        to be in pixels. X and Y are not defined by the documentation, so
        taking as a given that these are horizontal and vertical. N.B.
        the documentation states that the horizontal direction is fast."""

        beam_x, beam_y = struct.unpack(self._ii, self._tiff_header_bytes[1668:1676])[:2]
        pixel_x, pixel_y = struct.unpack(self._ii, self._tiff_header_bytes[1796:1804])[
            :2
        ]

        return beam_x * 1000 / pixel_x, beam_y * 1000 / pixel_y

    def _get_rayonix_detector_rotations(self):
        return check(FormatTIFFRayonix._get_rayonix_detector_rotations(self))

    def _get_rayonix_distance(self):
        return check(FormatTIFFRayonix._get_rayonix_distance(self))

    def _get_rayonix_pixel_size(self):
        return check(FormatTIFFRayonix._get_rayonix_pixel_size(self))

    def _get_rayonix_times(self):
        return check(FormatTIFFRayonix._get_rayonix_times(self))

    def _get_rayonix_timestamp(self):
        return check(FormatTIFFRayonix._get_rayonix_timestamp(self))

    def _get_rayonix_scan_angles(self):
        return check(FormatTIFFRayonix._get_rayonix_scan_angles(self))


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatTIFFRayonixXPP.understand(arg))



from __future__ import annotations

import dxtbx
from dxtbx.model.experiment_list import ExperimentList


def test_jf16M_matches_expected_hierarchy(dials_data):
    try:
        h5path = (
            dials_data("lysozyme_JF16M_4img", pathlib=True)
            / "lyso009a_0087.JF07T32V01_master_4img.h5"
        )
    except Exception as e:
        print(type(e), str(e))
        raise
    img = dxtbx.load(h5path)

    d1 = img.get_detector()

    expts_path = (
        dials_data("lysozyme_JF16M_4img", pathlib=True)
        / "lyso009a_0087.JF07T32V01_master_4img_imported.expt"
    )
    expts = ExperimentList.from_file(expts_path, check_format=False)

    d2 = expts[0].detector

    def recursive_test(pg1, pg2):
        assert pg1.is_similar_to(pg2)
        for c1, c2 in zip(pg1, pg2):
            recursive_test(c1, c2)

    recursive_test(d1.hierarchy(), d2.hierarchy())
